### HANDLING JSON FILES ###

import os
import ast
import json
from stpstone.utils.parsers.str import StrHandler
from typing import List, Dict, Any, Union


class JsonFiles:

    def dump_message(self, message, json_file):
        """
        DOCSTRING: DUMP MESSAGE AND SAVE TO SPECIFIC DIRECTORY IN THE NETWORK
        INPUTS: MESSAGE (STR IN DICT FORMAT), JSON FILE AS A COMPLETE FILE .JSON NAME
        OUTPUTS: WHETER FILE SAVING WAS SUCCESFUL OR NOT (OK, NOK)
        """
        with open(json_file, "w") as write_file:
            json.dump(message, write_file)
        if not os.path.exists(json_file):
            return 'NOK'
        else:
            return 'OK'

    def load_message(self, json_file, errors='ignore', encoding=None, decoding=None):
        """
        DOCSTRING: LOAD MESSAGE FROM JSON
        INPUTS: JSON FILE
        OUTPUTS: DATA IN MEMORY
        """
        if encoding is None:
            with open(json_file, "r", errors=errors) as read_file:
                return json.load(read_file)
        else:
            with open(json_file, errors=errors, encoding=encoding) as read_file:
                return json.loads(read_file.read().encode(encoding).decode(decoding))

    def loads_message_like(self, message):
        """
        DOCSTRING: LOAD A REPRESENTATION IN STR FROM A VALID MESSAGE TYPE, LIKE LIST, BYTE OR DICT
        INPUTS: MESSAGE
        OUTPUTS: JSON
        """
        return json.loads(message)

    def dict_to_json(self, message):
        """
        DOCSTRING: LOAD MESSAGE INTO JSON FORMAT
        INPUTS: MESSAGE
        OUTPUTS: JSON
        """
        return json.dumps(message)

    def send_json(self, message):
        """
        DOCSTRING: SEND JSON IN MEMORY, FOR INSTANCE IN AN API RESPONSE
        INPUTS: MESSAGE
        OUTPUTS: JSON IN MEMORY
        """
        return json.loads(json.dumps(message))

    def byte_to_json(self, byte_message):
        """
        DOCSTRING: DECRIPT BYTE FORMAT TO JSON
        INPUTS: BYTE MESSAGE
        OUTPUTS: JSON MESSAGE
        """
        jsonify_message = ast.literal_eval(StrHandler().get_between(
            str(byte_message), "b'", "'"))
        return JsonFiles().send_json(jsonify_message)

    def normalize_json_keys(self, json_:List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        DOCSTRING: VALIDATE THAT ALL DICTIONARIES IN THE LIST HAVE THE SAME KEYS. IF THEY DO,
            RETURN THE ORIGINAL LIST, OTHERWISE RETURN A NEW LIST WITH EMPTY VALUES FOR MISSING KEYS
        INPUTS: JSON_DATA (LIST[DICT[STR, ANY]]): A LIST OF DICTIONARIES TO VALIDATE
        OUTPUTS: LIST
        """
        # determine the set of all keys present in any of the dictionaries
        list_keys = set()
        for item in json_:
            list_keys.update(item.keys())
        # iterate over each dictionary and add missing keys with a value of 0
        for item in json_:
            for key in list_keys:
                if key not in item:
                    item[key] = 0
        return json_



### A CLASS TO CALCULATE CAPITAL REQUIREMENT 1 (CR1) FOR FINANCIAL INSTRUMENTS ###

from typing import Dict


class CR1Calculator:

    def __init__(self, float_ead: float, float_pd: float, float_lgd: float) -> None:
        """
        Args:
            float_ead (float): Exposure at Default.
            float_pd (float): Probability of Default.
            float_lgd (float): Loss Given Default.
        """
        self.float_ead = float_ead
        self.float_pd = float_pd
        self.float_lgd = float_lgd

    @property
    def calculate_k(self) -> float:
        """
        Calculates the capital requirement k, which is the product of the loss given default (LGD)
        and the probability of default (PD).

        Returns:
            float: The capital requirement k.
        """
        return self.float_lgd * self.float_pd

    @property
    def calculate_cr1(self) -> float:
        """
        Calculates the capital requirement 1 (CR1) by multiplying the exposure at default (EAD)
        by the capital requirement k, which is the product of the loss given default (LGD) and
        the probability of default (PD), and then by 12.5.

        Returns:
            float: The capital requirement 1 (CR1).
        """
        return 12.5 * self.calculate_k * self.float_ead

    @property
    def summary(self) -> Dict[str, float]:
        """
        A summary of all the values used to calculate the capital requirement 1 (CR1)

        Returns:
            Dict[str, float]: A dictionary containing the exposure at default (EAD),
            the probability of default (PD),
            the loss given default (LGD), the capital requirement k, and the capital
            requirement 1 (CR1).
        """
        return {
            'EAD': self.float_ead,
            'PD': self.float_pd,
            'LGD': self.float_lgd,
            'K': self.calculate_k,
            'CR1': self.calculate_cr1
        }



import pandas as pd
from math import nan
from datetime import datetime
from typing import Optional, List, Any, Tuple, Dict
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from selenium.webdriver.remote.webdriver import WebDriver
from stpstone._config.global_slots import YAML_YAHII_RATES
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.numbers import NumHandler
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD
from stpstone.utils.cals.handling_dates import DatesBR


class YahiiRatesBRMacro(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        int_delay_seconds: int = 20,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_YAHII_RATES,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs, 
            int_delay_seconds=int_delay_seconds,
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.int_delay_seconds = int_delay_seconds

    def td_value(self, str_value: str) -> float:
        """
        Get value from string

        Args:
            str_value (str): string to convert

        Returns:
            float: converted value
        """
        if str_value == "A/M":
            return "A/M"
        elif str_value == "–":
            return nan
        elif ("%" in str_value) and ("*" not in str_value):
            return float(str_value.replace("(-)", "-").replace("%", "").replace(".", "")\
                         .replace(",", ".").replace(" ", "")) / 100.0
        elif "(-)" in str_value or ("R$ " in str_value and len(str_value) <= 8):
            return float(str_value.replace("(-)", "-").replace("%", "").replace(".", "")\
                         .replace(",", ".").replace("R$ ", ""))
        elif str_value in ["", " "]:
            return ""
        elif NumHandler().is_numeric(str_value.replace(",", ".")) == True:
            return float(str_value.replace(",", "."))
        else:
            self.create_log.log_message(
                self.logger, 
                f"{str_value} of type {type(str_value)}, which cannot be handled by the method, "
                + f"please check", 
                "warning"
            )
            return ""

    def td_th_parser(self, source: str, list_th_td: List[Any]) -> List[Dict[str, Any]]:
        list_td_months = list()
        list_td_years = list()
        list_td_values = list()
        list_ser = list()
        # print(f"list_th_td: {list_th_td}")
        # populating list_td_months, list_td_years and list_td_values
        for td in list_th_td:
            if (td in YAML_YAHII_RATES["pmi_rf_rates"]["list_months_combined"] 
                and "/" not in td) \
                or "ACUMULADO" in td \
                or "ACUM" in td:
                list_td_months.append(td)
            elif len(td) == 4 and NumHandler().is_numeric(td) == True:
                try:
                    list_td_years.append(int(td))
                except ValueError:
                    list_td_values.append(self.td_value(td.replace(".", ",")))
            else:
                list_td_values.append(self.td_value(td))
        # cleaning list_td_months and list_td_years, regarding the source
        if source in ["igpm"]: 
            list_td_values.remove("A/M")
        elif source in ["ipcae"]:
            list_td_years = list_td_years[2:]
            list_td_months = ["JAN", "FEV", "MAR", "ACUM TRIM 1", "ABR", "MAI", "JUN", "ACUM TRIM 2", 
                              "JUL", "AGO", "SET", "ACUM TRIM 3", "OUT", "NOV", "DEZ", "ACUM TRIM 4", 
                              "ACUMULADO NO ANO"]
        elif source in ["ivar"]:
            list_td_months = [x for x in list_td_months if x not in ["JAN", "FEV", "MAR", "ABR", 
                                                                     "MAI", "JUN", "JUL", "AGO", 
                                                                     "SET", "OUT", "NOV", "DEZ"]]
        if source not in ["ipcae"]:
            list_td_months = ListHandler().remove_duplicates(list_td_months)
        list_td_years = ListHandler().remove_duplicates(list_td_years)
        list_td_years = [x for x in list_td_years \
                            if x <= DatesBR().year_number(DatesBR().curr_date) + 1 \
                            and x >= YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                                "int_year_min"]]
        list_td_months = [x for x in list_td_months if "ACUMULADONO" not in x]
        list_td_years.sort()
        # print(f"list_td_months: {list_td_months}")
        # print(f"list_td_years: {list_td_years}")
        # print(f"list_td_values: {[(i, x) for i, x in enumerate(list_td_values)]}")
        # populating list_ser with year, month and values
        for i_y, int_year in enumerate(list_td_years):
            if DatesBR().year_number(DatesBR().curr_date) < int_year: break
            i_m = 0
            if ("int_year_min" in YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]) \
                and ("int_cols_tbl" in YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]) \
                and ("int_step_tbl_begin" in YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][
                    source]):
                int_num_tbl = (int_year - YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][
                    source]["int_year_min"]) // YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][
                        source]["int_cols_tbl"]
                int_start_tbl = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_start"] \
                    + int_num_tbl * YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                        "int_step_tbl_begin"]
                int_cols_tbl = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_cols_tbl"]
                int_step_tbl_begin = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_step_tbl_begin"]
                int_year_min = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_year_min"]
                int_start_acc = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_start_acc"]
            else:
                int_num_tbl = 0
                int_cols_tbl = 0
                int_step_tbl_begin = 0
                int_year_min = 0
                int_start_acc = 0
                int_start_tbl = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                    "int_start"]
            for str_month in list_td_months:
                # position of the value in the list_td_values
                if "ACUMULADO" not in str_month:
                    int_pos = int_start_tbl \
                        + YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]["int_y"] * i_y \
                        + YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]["int_m"] * i_m
                else:
                    int_pos = YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                        "int_start_acc"] \
                        + YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]["int_y"] * i_y
                if source in ["tr"] and int_num_tbl > 1 and "ACUM" not in str_month:
                    int_pos -= int(int_num_tbl - 1)
                if source in ["tr"] and int_num_tbl > 0 and "ACUM" in str_month:
                    int_pos -= int_num_tbl
                # populating serialized list
                try:
                    list_ser.append({
                        "INT_START": YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source][
                            "int_start"],
                        "INT_Y": YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]["int_y"],
                        "INT_M": YAML_YAHII_RATES["pmi_rf_rates"]["dict_fw_td_th"][source]["int_m"],
                        "INT_COLS_TBL": int_cols_tbl,
                        "INT_STEP_TBL_BEGIN": int_step_tbl_begin,
                        "INT_YEAR_MIN": int_year_min,
                        "INT_START_ACC": int_start_acc,
                        "INT_NUM_TBL": int_num_tbl,
                        "I_Y": i_y,
                        "I_M": i_m,
                        "INT_START_TBL": int_start_tbl,
                        "INDEX_TD": int_pos,
                        "YEAR": int_year,
                        "MONTH": str_month,
                        "VALUE": float(list_td_values[int_pos]) / 100.0 \
                            if (source in ["ipcae"] and "ACUMULADO NO ANO" not in str_month 
                                and list_td_values[int_pos] != "") \
                            else list_td_values[int_pos],
                        "ECONOMIC_INDICATOR": "CDI" if source == "cetip" else source.upper()
                    })
                    i_m += 1
                except IndexError:
                    break
        return list_ser
        
    def list_web_elements(self, cls_selenium_wd: SeleniumWD, web_driver: WebDriver, xpath_: str) \
        -> List[Any]:
        """
        Get list of web elements from xpath

        Args:
            cls_selenium_wd (SeleniumWD): selenium web web_driver
            web_driver (WebDriver): selenium web web_driver
            xpath (str): xpath to find elements

        Returns:
            list: list of contents of web elements
        """
        list_els = cls_selenium_wd.find_elements(web_driver, xpath_)
        list_els = [x.text for x in list_els]
        return list_els

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        try:
            source = self.get_query_params(resp_req.url, "source").lower()
            cls_selenium_wd = SeleniumWD(resp_req.url, bl_headless=True, bl_incognito=True)
            web_driver = cls_selenium_wd.get_web_driver
            list_th_td = self.list_web_elements(
                cls_selenium_wd, web_driver, YAML_YAHII_RATES["pmi_rf_rates"]['xpaths']['list_th_td'])
            list_ser = self.td_th_parser(source, list_th_td)
        finally:
            web_driver.quit()
        return pd.DataFrame(list_ser)




from collections import deque
from typing import Any


class Deque:

    def __init__(self) -> None:
        self._items: deque[Any] = deque()

    @property
    def is_empty(self) -> bool:
        return not self._items

    def add_front(self, item: Any) -> None:
        self._items.append(item)

    def add_rear(self, item: Any) -> None:
        self._items.appendleft(item)

    @property
    def remove_front(self) -> Any:
        return self._items.pop()

    @property
    def remove_rear(self) -> Any:
        return self._items.popleft()

    @property
    def size(self) -> int:
        return len(self._items)



### METADATA: https://ranaroussi.github.io/yfinance/index.html ###


# pypi.org libs
import pandas as pd
from yfinance import download
from typing import List, Optional
from datetime import datetime
# local libs
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.utils.cals.handling_dates import DatesBR


class YFinanceWS(metaclass=TypeChecker):

    def __init__(
        self,
        list_tickers:List[str],
        dt_start:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 52).strftime('%Y-%m-%d'),
        dt_end:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1).strftime('%Y-%m-%d'),
        session:Optional[ProxyScrapeAll]=None
    ) -> None:
        self.list_tickers = list_tickers
        self.session = session
        self.dt_start = dt_start
        self.dt_end = dt_end

    @property
    def mktdata(self) -> pd.DataFrame:
        df_ = download(
            tickers=self.list_tickers,
            start=self.dt_start.strftime('%Y-%m-%d'),
            end=self.dt_end.strftime('%Y-%m-%d'),
            group_by='ticker',
            session=self.session
        )
        return df_



import pandas as pd
from datetime import datetime
from logging import Logger
from time import sleep
from typing import Any, List, Optional, Tuple
from requests import Response
from sqlalchemy.orm import Session
from stpstone._config.global_slots import YAML_B3_FUTURES_CLOSING_ADJ
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.str import StrHandler


class FuturesClosingAdjB3(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
    ) -> None:
        super().__init__(
            dict_metadata=YAML_B3_FUTURES_CLOSING_ADJ,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs,
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.repr_dt_ref = self.dt_ref.strftime("%d/%m/%Y").replace("/", "%2F")

    def td_th_parser(
        self, resp_req: Response, list_th: List[Any]
    ) -> Tuple[List[Any], int, Optional[int]]:
        list_headers = list_th.copy()
        int_init_td = 0
        int_end_td = None
        # for using this workaround, please pass a dummy variable to the url, within the YAML file,
        #   like https://example.com/app/#source=dummy_1&bl_debug=True
        return list_headers, int_init_td, int_end_td

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        bl_debug = (
            True
            if StrHandler().match_string_like(resp_req.url, "*bl_debug=True*") == True
            else False
        )
        root = HtmlHandler().lxml_parser(resp_req)
        # export html tree to data folder, if is user's will
        if bl_debug == True:
            path_project = DirFilesManagement().find_project_root(
                marker="pyproject.toml"
            )
            HtmlHandler().html_tree(root, file_path=rf"{path_project}/data/test.html")
        list_th = [
            x.text.strip().replace("R$)", "BRL")
            for x in HtmlHandler().lxml_xpath(
                root,
                YAML_B3_FUTURES_CLOSING_ADJ["futures_closing_adj"]["xpaths"]["list_th"],
            )
        ]
        list_td = [
            (
                ""
                if x.text is None
                else x.text.replace("\xa0", "")
                .replace(".", "")
                .replace(",", ".")
                .strip()
            )
            for x in HtmlHandler().lxml_xpath(
                root,
                YAML_B3_FUTURES_CLOSING_ADJ["futures_closing_adj"]["xpaths"]["list_td"],
            )
        ]
        # deal with data/headers specificity for the project
        list_headers, int_init_td, int_end_td = self.td_th_parser(resp_req, list_th)
        if bl_debug == True:
            print(list_headers)
            print(f"LEN LIST HEADERS: {len(list_headers)}")
            print(list_td[int_init_td:int_end_td])
            print(f"LEN LIST TD: {len(list_td[int_init_td:int_end_td])}")
        list_ser = HandlingDicts().pair_headers_with_data(
            list_headers, list_td[int_init_td:int_end_td]
        )
        return pd.DataFrame(list_ser)



### GENERIC PLUGING FOR AIRFLOW ###

# pypi.org libs
from typing import Dict, Any
# project modules
from stpstone.utils.cals.handling_dates import DatesBR


class AirflowPlugins:

    def validate_working_day(self, **kwargs: Dict[str, Any]) -> None:
        """
        Validates whether the provided date (`kwargs['ds']`) is a working day, otherwise stop the DAG.

        Args:
            kwargs (dict): Airflow context dictionary. Must include 'ds' key.

        Returns:
            None
        """
        ti = kwargs['ti']
        bl_workng_day = DatesBR().is_working_day(kwargs['ds'])
        ti.xcom_push(key='bl_continue', value=bl_workng_day)



import redis
from typing import Dict
from __future__ import annotations


class RedisClient:
    """
    DOCSTRING: SINGLETON REDIS CLIENT FOR MANAGING CONNECTIONS
    INPUTS: -
    OUTPUTS: -
    """
    _instance = None

    def __init__(self, str_host:str='localhost', int_port:int=6379, bl_decode_resp:bool=True) -> None:
        """
        DOCSTRING: STORE THE CONNECTION CONFIGURATION
        INPUTS: HOST:STR, PORT:INT, DECODE_RESP:BOOL
        OUTPUTS: -
        """
        self.str_host = str_host
        self.int_port = int_port
        self.bl_decode_resp = bl_decode_resp

    def __new__(cls, *args, **kwargs) -> RedisClient:
        """
        DOCSTRING: NEW CONNECTION
        INPUTS: CREATE OR RETRIEVE THE SINGLETON INSTANCE OF REDISCLIENT
        OUTPUTS: CLASS INSTANCE
        """
        if not cls._instance:
            #   create a new instance if it doesn't exist
            cls._instance = super(RedisClient, cls).__new__(cls)
            #   initialize the redis client connection
            cls._instance._redis_client = cls._connect(*args, **kwargs)
        return cls._instance

    @staticmethod
    def _load_config(str_host:str, int_port:int, bl_decode_resp:bool) -> Dict[str, object]:
        """
        DOCSTRING: LOAD REDIS CONFIGURATION PARAMETERS
        INPUTS:
            - str_host: REDIS HOST
            - int_port: REDIS PORT
            - bl_decode_resp: WHETHER TO DECODE RESPONSES
        OUTPUTS: CONFIGURATION DICTIONARY
        """
        return {
            'host': str_host,
            'port': int_port,
            'decode_responses': bl_decode_resp
        }

    @classmethod
    def _connect(cls, str_host:str, int_port:int, bl_decode_resp:bool) -> redis.StrictRedis:
        """
        DOCSTRING: ESTABLISH A CONNECTION TO REDIS
        INPUTS:
            - str_host: REDIS HOST
            - int_port: REDIS PORT
            - bl_decode_resp: WHETHER TO DECODE RESPONSES
        OUTPUTS: REDIS CLIENT INSTANCE
        """
        # load configuration parameters
        config = cls._load_config(str_host, int_port, bl_decode_resp)
        # create and return a redis client instance
        return redis.StrictRedis(**config)

    @classmethod
    def get(cls) -> redis.StrictRedis:
        """
        DOCSTRING: RETRIEVE THE REDIS CLIENT INSTANCE
        INPUTS: -
        OUTPUTS: REDIS CLIENT INSTANCE
        """
        # return the singleton redis client instance
        return cls()._redis_client



import pandas as pd
from logging import Logger, Tuple
from numbers import Number
from typing import Dict, List, Optional
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.pipelines.generic import generic_pipeline


class DataFrameValidator:

    def __init__(self, df_: pd.DataFrame, logger: Logger):
        self.df_ = df_.copy()
        self.logger = logger
        self.create_log = CreateLog()

    @property
    def check_missing_values(self):
        missing_summary = self.df_.isnull().sum()
        self.create_log.warning(
            self.logger, "Missing Values:\n", missing_summary[missing_summary > 0]
        )

    @property
    def check_duplicates(self):
        duplicate_count = self.df_.duplicated().sum()
        if duplicate_count:
            self.create_log.warning(
                self.logger, f"Found {duplicate_count} duplicate rows."
            )

    def validate_ranges(
        self, dict_rng_constraints: Dict[str, Tuple[Number, Number]]
    ) -> None:
        for col, (low, high) in dict_rng_constraints.items():
            out_of_bounds = self.df_[(self.df_[col] < low) | (self.df_[col] > high)]
            if not out_of_bounds.empty:
                self.create_log.warning(
                    self.logger,
                    f"{len(out_of_bounds)} values in '{col}' out of range ({low} - {high})",
                )

    def validate_dates(self, col_start: str, col_sup: str) -> None:
        list_inv_dates = self.df_[self.df_[col_start] > self.df_[col_sup]]
        if not list_inv_dates.empty:
            self.create_log.warning(
                self.logger,
                f"Found {len(list_inv_dates)} rows where {col_start} is after {col_sup}.",
            )

    def validate_categorical_values(
        self, col_: Optional[str] = None, list_allowed_values: Optional[str] = None
    ):
        list_inv_values = self.df_[~self.df_[col_].isin(list_allowed_values)]
        if not list_inv_values.empty:
            self.create_log.warning(
                self.logger,
                f"Found invalid values in '{col_}':",
                list_inv_values[col_].unique(),
            )

    def pipeline(
        self,
        dict_rng_constraints: Dict[str, Tuple[Number, Number]] = None,
        col_start: Optional[str] = None,
        col_sup: Optional[str] = None,
        list_tup_categorical_constraints: Optional[Tuple[str, List[str]]] = None,
    ):
        steps = [
            self.check_missing_values,
            self.check_duplicates,
        ]
        if dict_rng_constraints:
            steps.append(self.validate_ranges(dict_rng_constraints))
        if col_start and col_sup:
            steps.append(self.validate_dates(col_start, col_sup))
        if list_tup_categorical_constraints:
            col_, allowed_values = list_tup_categorical_constraints
            steps.append(self.validate_categorical_values(col_, allowed_values))
        return generic_pipeline(self.df_, steps)



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class BinaryDivider(metaclass=TypeChecker):

    def __init__(self, dividend: int, divisor: int) -> None:
        """
        Initialize the Binary Divider with two binary inputs.

        Args:
            dividend (int): The dividend, the number being divided.
            divisor (int): The divisor, the number by which we are dividing.

        Returns:
            None

        Raises:
            ValueError: If the divisor is zero.
        """
        if divisor == 0:
            raise ValueError("Divisor cannot be zero.")
        self.dividend = dividend
        self.divisor = divisor

    @property
    def divide(self) -> tuple:
        """
        Divide the dividend by the divisor using binary division.

        The algorithm works by shifting the divisor left and subtracting it
        from the dividend if the dividend is greater than or equal to the
        divisor. The result is then shifted back to the right, and the
        remainder is returned.

        Args:
            None

        Returns:
            tuple: (quotient, remainder)

        Raises:
            ValueError: If the divisor is zero.
        """
        quotient = 0
        remainder = self.dividend
        for i in range(7, -1, -1):
            # if remainder is greater than or equal to divisor, subtract divisor
            if remainder >= (self.divisor << i):
                remainder -= (self.divisor << i)
                # set the i-th bit of the quotient
                quotient |= (1 << i)
        return quotient, remainder


if __name__ == "__main__":
    divider = BinaryDivider(0b1100, 0b0010)
    quotient, remainder = divider.divide()
    # output: Quotient and Remainder in binary
    print(f"Binary Divider (1100 / 10) => Quotient: {bin(quotient)}, Remainder: {bin(remainder)}")




### PRICING FORWARD CONTRACTS ###

class ForwardBR:

    def forward_contract_pricing(self, spot, bid_forward_contract, rate_cost_period, leverage,
                                 number_contracts):
        """
        DOCSTRING: FORWARD CONTRACT FEATURES
        INPUTS: SPOT PRICE, BID AND RATE COST FOR THE PERIOD
        OUTPUTS: DICTIONARY (MTM, PCT RETURN AND CONTRACT NOTIONAL)
        """
        return {
            'mtm': (float(spot) - float(bid_forward_contract) * (
                1.0 + rate_cost_period)) * float(leverage) * float(number_contracts),
            'pct_retun': (float(spot) - float(bid_forward_contract) * (
                1.0 + rate_cost_period)) * float(leverage) / float(bid_forward_contract),
            'notional': float(spot) * float(leverage) * float(number_contracts)
        }



from requests import request
from typing import Dict, Union, List, Optional
from logging import Logger
from stpstone.utils.connections.netops.proxies.abc import ABCSession


class ProxyScrapeAll(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        resp_req = request(
            "GET",
            "https://api.proxyscrape.com/v4/free-proxy-list/get?request=display_proxies&proxy_format=protocolipport&format=json",
        )
        resp_req.raise_for_status()
        json_proxies = resp_req.json()
        return [
            {
                "protocol": str(dict_["protocol"]).lower(),
                "bl_alive": bool(dict_["alive"]),
                "status": str(dict_["ip_data"]["status"]) if "ip_data" in dict_ else "",
                "alive_since": float(dict_["alive_since"]),
                "anonymity": str(dict_["anonymity"]).lower(),
                "average_timeout": float(dict_["average_timeout"]),
                "first_seen": float(dict_["first_seen"]),
                "ip_data": str(dict_["ip_data"]["as"]) if "ip_data" in dict_ else "",
                "ip_name": str(dict_["ip_data"]["asname"]) if "ip_data" in dict_ else "",
                "timezone": str(dict_["ip_data"]["timezone"]) if "ip_data" in dict_ else "",
                "continent": str(dict_["ip_data"]["continent"]) if "ip_data" in dict_ else "",
                "continent_code": str(dict_["ip_data"]["continentCode"]) if "ip_data" in dict_ else "",
                "country": str(dict_["ip_data"]["country"]) if "ip_data" in dict_ else "",
                "country_code": str(dict_["ip_data"]["countryCode"]) if "ip_data" in dict_ else "",
                "city": str(dict_["ip_data"]["city"]) if "ip_data" in dict_ else "",
                "district": str(dict_["ip_data"]["district"]) if "ip_data" in dict_ else "",
                "region_name": str(dict_["ip_data"]["regionName"]) if "ip_data" in dict_ else "",
                "zip": str(dict_["ip_data"]["zip"]) if "ip_data" in dict_ else "",
                "bl_hosting": bool(dict_["ip_data"]["hosting"]) if "ip_data" in dict_ else "",
                "isp": str(dict_["ip_data"]["isp"]) if "ip_data" in dict_ else "",
                "latitude": float(dict_["ip_data"]["lat"]) if "ip_data" in dict_ else "",
                "longitude": float(dict_["ip_data"]["lon"]) if "ip_data" in dict_ else "",
                "organization": str(dict_["ip_data"]["org"]) if "ip_data" in dict_ else "",
                "proxy": str(dict_["proxy"]),
                "ip": str(dict_["ip"]),
                "port": int(dict_["port"]),
                "bl_ssl": bool(dict_["ssl"]),
                "timeout": float(dict_["timeout"]),
                "times_alive": float(dict_["times_alive"]),
                "times_dead": float(dict_["times_dead"]),
                "ratio_times_alive_dead": float(dict_["times_alive"] / dict_["times_dead"])
                    if "times_alive" in dict_ and "times_dead" in dict_ and dict_["times_dead"] != 0
                    else 0,
                "uptime": float(dict_["uptime"])
            } for dict_ in json_proxies["proxies"]
        ]


class ProxyScrapeCountry(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        resp_req = request(
            "GET",
            f"https://api.proxyscrape.com/v4/free-proxy-list/get?request=get_proxies&country={self.str_country_code.lower()}&skip=0&proxy_format=protocolipport&format=json&limit=1000",
        )
        resp_req.raise_for_status()
        json_proxies = resp_req.json()
        return [
            {
                "protocol": str(dict_["protocol"]).lower(),
                "bl_alive": bool(dict_["alive"]),
                "status": str(dict_["ip_data"]["status"]) if "ip_data" in dict_ else "",
                "alive_since": float(dict_["alive_since"]),
                "anonymity": str(dict_["anonymity"]).lower(),
                "average_timeout": float(dict_["average_timeout"]),
                "first_seen": float(dict_["first_seen"]),
                "ip_data": str(dict_["ip_data"]["as"]) if "ip_data" in dict_ else "",
                "ip_name": str(dict_["ip_data"]["asname"]) if "ip_data" in dict_ else "",
                "timezone": str(dict_["ip_data"]["timezone"]) if "ip_data" in dict_ else "",
                "continent": str(dict_["ip_data"]["continent"]) if "ip_data" in dict_ else "",
                "continent_code": str(dict_["ip_data"]["continentCode"]) if "ip_data" in dict_ else "",
                "country": str(dict_["ip_data"]["country"]) if "ip_data" in dict_ else "",
                "country_code": str(dict_["ip_data"]["countryCode"]) if "ip_data" in dict_ else "",
                "city": str(dict_["ip_data"]["city"]) if "ip_data" in dict_ else "",
                "district": str(dict_["ip_data"]["district"]) if "ip_data" in dict_ else "",
                "region_name": str(dict_["ip_data"]["regionName"]) if "ip_data" in dict_ else "",
                "zip": str(dict_["ip_data"]["zip"]) if "ip_data" in dict_ else "",
                "bl_hosting": bool(dict_["ip_data"]["hosting"]) if "ip_data" in dict_ else "",
                "isp": str(dict_["ip_data"]["isp"]) if "ip_data" in dict_ else "",
                "latitude": float(dict_["ip_data"]["lat"]) if "ip_data" in dict_ else "",
                "longitude": float(dict_["ip_data"]["lon"]) if "ip_data" in dict_ else "",
                "organization": str(dict_["ip_data"]["org"]) if "ip_data" in dict_ else "",
                "proxy": str(dict_["proxy"]),
                "ip": str(dict_["ip"]),
                "port": int(dict_["port"]),
                "bl_ssl": bool(dict_["ssl"]),
                "timeout": float(dict_["timeout"]),
                "times_alive": float(dict_["times_alive"]),
                "times_dead": float(dict_["times_dead"]),
                "ratio_times_alive_dead": float(dict_["times_alive"] / dict_["times_dead"])
                    if "times_alive" in dict_ and "times_dead" in dict_ and dict_["times_dead"] != 0
                    else 0,
                "uptime": float(dict_["uptime"])
            } for dict_ in json_proxies["proxies"]
        ]



from stpstone.dsa.stacks.simple_stack import Stack
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from typing import Optional


class BaseConverter(metaclass=TypeChecker):

    def __init__(
        self,
        str_num: str,
        int_from_base: Optional[int] = 10,
        int_to_base: Optional[int] = 2
    ) -> None:
        """
        Initialize the BaseConverter with a str_num and its conversion bases.

        Args:
            str_num (str): The str_num to be converted, represented as a string.
            int_from_base (int): The base of the input str_num, must be between 2 and 16.
            int_to_base (int): The base to convert the input str_num to, must be between 2 and 16.

        Raises:
            ValueError: If either `int_from_base` or `int_to_base` is not between 2 and 16.
        """
        if not (2 <= int_from_base <= 16) or not (2 <= int_to_base <= 16):
            raise ValueError("Base must be between 2 and 16")
        self.str_num = str_num
        self.int_from_base = int_from_base
        self.int_to_base = int_to_base
        self.digits = "0123456789ABCDEF"
        for char in self.str_num:
            if char.upper() not in self.digits:
                raise ValueError(f"Invalid str_num: {str_num}, please enter a valid "
                                 + f"str_num with characters between {self.digits}")

    @property
    def _to_decimal(self) -> int:
        decimal_value = 0
        for i, digit in enumerate(reversed(self.str_num)):
            decimal_value += self.digits.index(digit.upper()) * (self.int_from_base ** i)
        return decimal_value

    @property
    def convert(self) -> str:
        decimal_value = self._to_decimal
        if self.int_to_base == 10:
            return str(decimal_value)
        if decimal_value == 0:
            return "0"
        rem_stack = Stack()
        num = decimal_value
        while num > 0:
            rem_stack.push(num % self.int_to_base)
            num //= self.int_to_base
        new_string = ""
        while not rem_stack.is_empty:
            new_string += self.digits[rem_stack.pop()]
        return new_string



import socket
import requests
import psutil


class NetworkInfo:
    """
    A class to retrieve network-related information such as IP addresses, port status, and open ports.
    """

    @property
    def get_public_ip_address(self):
        """
        Retrieves the public IP address of the machine using an external service.
        Returns:
            str: The public IP address, or an error message if the request fails.
        """
        try:
            response = requests.get('https://api.ipify.org?format=json')
            response.raise_for_status()
            public_ip = response.json()['ip']
            return public_ip
        except requests.RequestException as e:
            return f'Error fetching public IP: {e}'

    def is_port_in_use(self, port):
        """
        Checks if a specific port is in use on the local machine.

        Args:
            port (int): The port number to check.

        Returns:
            bool: True if the port is in use, False otherwise.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                #   try to bind to the port
                s.bind(('0.0.0.0', port))
                #   port is available
                return False
            except OSError:
                #   port is in use
                return True

    @property
    def get_available_port(self):
        """
        Finds and returns an available port on the local machine.

        Returns:
            int: An available port number.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            #   bind to an available port (0 means OS assigns a free port)
            s.bind(('0.0.0.0', 0))
            #   get the assigned port number
            port = s.getsockname()[1]
            return port

    @property
    def get_open_ports(self):
        """
        Retrieves a list of all open ports on the local machine.

        Returns:
            list: A list of open port numbers.
        """
        open_ports = []
        for conn in psutil.net_connections():
            if conn.status == psutil.CONN_LISTEN and conn.laddr:
                open_ports.append(conn.laddr.port)
        return open_ports


if __name__ == '__main__':
    network_info = NetworkInfo()

    # get public IP address
    public_ip = network_info.get_public_ip_address
    print(f'Public IP Address: {public_ip}')

    # check if a port is in use
    port_to_check = 8080
    if network_info.is_port_in_use(port_to_check):
        print(f'Port {port_to_check} is in use.')
    else:
        print(f'Port {port_to_check} is available.')

    # get an available port
    available_port = network_info.get_available_port
    print(f'Available Port: {available_port}')

    # get all opened ports
    open_ports = network_info.get_open_ports
    print(f'Open Ports: {open_ports}')



import backoff
import requests
import random
from typing import List
from lxml import html
from requests.exceptions import (ReadTimeout, ConnectTimeout, ChunkedEncodingError,
                                 RequestException, HTTPError)


class UserAgents:

    @property
    @backoff.on_exception(
        backoff.expo,
        (RequestException, HTTPError, ReadTimeout, ConnectTimeout, ChunkedEncodingError),
        max_tries=20,
        base=2,
        factor=2,
        max_value=1200
    )
    def fetch_user_agents(self) -> List[str]:
        list_ = list()
        url = "https://gist.github.com/pzb/b4b6f57144aea7827ae4"
        xpath = '//*[@id="file-user-agents-txt-LC{}"]/text()'
        i = 1
        dict_headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
        }
        resp_req = requests.get(url, headers=dict_headers, timeout=10)
        resp_req.raise_for_status()
        tree = html.fromstring(resp_req.content)
        while True:
            agent = tree.xpath(xpath.format(i))
            if not agent:
                break
            list_.append(agent[0].strip())
            i += 1
        return list_

    @property
    def get_random_user_agent(self):
        fallback_agents = [
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/119.0",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/119.0",
        ]
        list_ = self.fetch_user_agents
        if list_ and len(list_) > 0:
            random_index = random.randint(0, len(list_) - 1)
            return list_[random_index]
        else:
            return random.choice(fallback_agents)



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple, Dict, Union
from lxml.html import HtmlElement
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from numbers import Number
from stpstone._config.global_slots import YAML_US_SLICKCHARTS_INDEXES_COMPONENTS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.numbers import NumHandler


class SlickChartsIndexesComponents(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_US_SLICKCHARTS_INDEXES_COMPONENTS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs

    def td_th_parser(self, root: HtmlElement, resp_req: Response) -> pd.DataFrame:
        list_ser = []
        source = self.get_query_params(resp_req.url, "source")
        for i in range(1, 600):
            try:
                el_tr = HtmlHandler().lxml_xpath(
                    root,
                    YAML_US_SLICKCHARTS_INDEXES_COMPONENTS[source]["xpaths"]["list_tr"].format(i))[0]
            except IndexError:
                break
            list_ser.append({
                "NUM_COMPANY": HtmlHandler().lxml_xpath(el_tr, "./td[1]")[0].text.strip(),
                "NAME_COMPANY": HtmlHandler().lxml_xpath(el_tr, "./td[2]/a")[0].text.strip(),
                "TICKER": HtmlHandler().lxml_xpath(el_tr, "./td[3]/a")[0].text.strip(),
                "WEIGHT": float(HtmlHandler().lxml_xpath(el_tr, "./td[4]")[0].text.strip()\
                    .replace("%", ""))/100.0,
            })
        return pd.DataFrame(list_ser)

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        bl_debug = True if StrHandler().match_string_like(
            resp_req.url, "*bl_debug=True*") == True else False
        root = HtmlHandler().lxml_parser(resp_req)
        # export html tree to data folder, if is user's will
        if bl_debug == True:
            path_project = DirFilesManagement().find_project_root(marker="pyproject.toml")
            HtmlHandler().html_tree(root, file_path=rf"{path_project}/data/test.html")
        return pd.DataFrame(self.td_th_parser(root, resp_req))



### CONDITIONAL PIPELINES ###

from typing import Any, List


def conditionalpipeline(data: Any, functions: List[tuple]) -> Any:
    """
    Applies functions conditionally based on a predicate
    Examples of usage:
        Fraud detection, rule-based processing
    Args:
        data (Any): Initial input
        functions (List[tuple]): List of (predicate, function) tuples
    Returns:
        Any: Processed data
    """
    for condition, func in functions:
        if condition(data):
            data = func(data)
    return data


if __name__ == "__main__":

    def is_even(x):
        return x % 2 == 0

    def double(x):
        return x * 2

    def triple(x):
        return x * 3

    steps = [(is_even, double), (lambda x: x > 10, triple)]
    result = conditionalpipeline(6, steps)
    print(result)



from typing import Literal
from stpstone.dsa.stacks.simple_stack import Stack
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class ExpressionConverter(metaclass=TypeChecker):

    def __init__(
        self,
        str_expr: str,
        str_from_type: Literal["infix", "postfix", "prefix"],
        str_to_type: Literal["infix", "postfix", "prefix"]
    ) -> None:
        self.str_expr = str_expr
        self.str_from_type = str_from_type
        self.str_to_type = str_to_type
        self.prec = {}
        self.prec["*"] = 3
        self.prec["/"] = 3
        self.prec["+"] = 2
        self.prec["-"] = 2
        self.prec["("] = 1
        self.postfix_list = []
        self.token_list = str_expr.upper().split(" ")
        self.str_operands = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        self.str_operators = "+-*/()"

    @property
    def _infix_to_postfix(self) -> str:
        cls_stack = Stack()
        for token in self.token_list:
            if token in self.str_operands:
                self.postfix_list.append(token)
            elif token == "(":
                cls_stack.push(token)
            elif token == ")":
                top_token = cls_stack.pop()
                while top_token != "(":
                    self.postfix_list.append(top_token)
                    top_token = cls_stack.pop()
            else:
                while (not cls_stack.is_empty) and (self.prec[cls_stack.peek] >= self.prec[token]):
                    self.postfix_list.append(cls_stack.pop())
                cls_stack.push(token)
        while not cls_stack.is_empty:
            self.postfix_list.append(cls_stack.pop())
        return " ".join(self.postfix_list)

    @property
    def _infix_to_prefix(self) -> str:
        postfix_expr = self._infix_to_postfix
        cls_stack = Stack()
        for token in postfix_expr.split(" "):
            if token in self.str_operands:
                cls_stack.push(token)
            else:
                right = cls_stack.pop()
                left = cls_stack.pop()
                expr = f"{token} {left} {right}"
                cls_stack.push(expr)
        return cls_stack.pop()

    @property
    def _postfix_to_infix(self) -> str:
        cls_stack = Stack()
        for token in self.token_list:
            if token in self.str_operands:
                cls_stack.push(token)
            else:
                right = cls_stack.pop()
                left = cls_stack.pop()
                expr = f"({left} {token} {right})"
                cls_stack.push(expr)
        return cls_stack.pop()

    @property
    def _postfix_to_prefix(self) -> str:
        cls_stack = Stack()
        for token in self.token_list:
            if token in self.str_operands:
                cls_stack.push(token)
            else:
                right = cls_stack.pop()
                left = cls_stack.pop()
                expr = f"{token} {left} {right}"
                cls_stack.push(expr)
        return cls_stack.pop()

    @property
    def _prefix_to_infix(self) -> str:
        cls_stack = Stack()
        for token in reversed(self.token_list):
            if token in self.str_operands:
                cls_stack.push(token)
            else:
                left = cls_stack.pop()
                right = cls_stack.pop()
                expr = f"({left} {token} {right})"
                cls_stack.push(expr)
        return cls_stack.pop()

    @property
    def _prefix_to_postfix(self) -> str:
        cls_stack = Stack()
        for token in reversed(self.token_list):
            if token in self.str_operands:
                cls_stack.push(token)
            else:
                left = cls_stack.pop()
                right = cls_stack.pop()
                expr = f"{left} {right} {token}"
                cls_stack.push(expr)
        return cls_stack.pop()

    @property
    def convert(self) -> str:
        if self.str_from_type == "infix" and self.str_to_type == "postfix":
            return self._infix_to_postfix
        elif self.str_from_type == "infix" and self.str_to_type == "prefix":
            return self._infix_to_prefix
        elif self.str_from_type == "postfix" and self.str_to_type == "infix":
            return self._postfix_to_infix
        elif self.str_from_type == "postfix" and self.str_to_type == "prefix":
            return self._postfix_to_prefix
        elif self.str_from_type == "prefix" and self.str_to_type == "infix":
            return self._prefix_to_infix
        elif self.str_from_type == "prefix" and self.str_to_type == "postfix":
            return self._prefix_to_postfix
        else:
            raise ValueError("Invalid conversion type")



from getpass import getuser
from logging import Logger
from typing import Optional
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.loggs.create_logs import CreateLog


def iniciating_logging(logger:Logger, path_log:Optional[str]=None) -> None:
    if path_log != None:
        dispatch = CreateLog().creating_parent_folder(path_log)
        CreateLog().info(logger, 'Logs parent directory: {}'.format(path_log))
        if dispatch == True:
            CreateLog().info(logger, 'Logs parent directory created successfully.')
        elif dispatch == False:
            CreateLog().info(
                logger, 'Logs parent directory could not be created.')
        else:
            raise Exception(
                'Unexpected dispatch value: {}'.format(dispatch))
    CreateLog().info(logger, 'Rotina iniciada em {}'.format(
        str(DatesBR().curr_date_time())))
    CreateLog().info(logger, 'Operador da rotina {}'.format(str(getuser())))



from typing import List, Optional
from openai import OpenAI
from openai.api_resources.chat_completion import ChatCompletion
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone._config.global_slots import YAML_LLMS


class GPT():

    def __init__(self, api_key:str, str_model:str, int_max_tokens:int=100,
                 str_context:Optional[str]=None, bl_stream:bool=False) -> None:
        """Initialize the GPT class.

        Parameters
        ----------
        api_key : str
            The API key for accessing the OpenAI service.
        str_model : str
            The model name to use for generating completions.
        int_max_tokens : int, optional
            Maximum number of tokens for the completion. Defaults to 100.
        str_context : Optional[str], optional
            Optional context to provide to the model. Defaults to None.
        bl_stream : bool, optional
            Whether to stream the completion. Defaults to False.

        Returns
        -------
        None

        Notes
        -----
        This class provides an interface to OpenAI's GPT models for text generation.

        Documentation: https://platform.openai.com/docs/guides/gpt
        Models available: https://platform.openai.com/docs/models/gpt

        When choosing a model, consider:
        - Model capabilities (some support longer contexts)
        - Cost (pricing varies by model)
        - Token limits
        """

        self.api_key = api_key
        self.str_model = str_model
        self.int_max_tokens = int_max_tokens
        self.str_context = str_context
        self.bl_strem=bl_stream,
        self.client = OpenAI(api_key=self.api_key)

    def run_prompt(self, list_tuple:List[tuple]) -> ChatCompletion:
        """Run the prompt on the model.

        Parameters
        ----------
        list_tuple : List[tuple]
            List of tuples with the information to build the prompt.

        Returns
        -------
        ChatCompletion
            The response from the model.

        Notes
        -----
        The list of tuples must have the following structure:
        - Each tuple must have two elements: type and content.
        - The type must be one of the following: 'text' or 'image_url'.
        - The content must be a string with the content of the prompt.

        The context, if provided, will be added to the prompt as a system message.
        """
        list_ = list()
        # user's information in order to build the prompt
        dict_content = {
            "role": "user"
        }
        # looping within types and messages, in order to create the content of the prompt
        for tup_ in list_tuple:
            if tup_[0] == "text":
                list_.append({
                    "type": str(tup_[0]).lower(),
                    str(tup_[0]).lower(): str(tup_[1])
                })
            elif tup_[0] == "image_url":
                list_.append({
                    "type": str(tup_[0]).lower(),
                    "image_url": {
                        "url": str(tup_[1]).lower()
                    }
                })
        # creating the message content
        dict_content = HandlingDicts().merge_n_dicts(dict_content, {"content": list_})
        # add context info, regarding the prompt
        if self.str_context is not None:
            list_prompt = [
                {
                    "role": "system",
                    "content": self.str_context
                },
                dict_content
            ]
        else:
            list_prompt = [dict_content]
        # request llm info
        return self.client.chat.completions.create(
            model=self.str_model,
            messages=list_prompt,
            max_tokens=self.int_max_tokens,
            strem=self.bl_strem
        )



from stpstone.dsa.stacks.simple_stack import Stack


class BalanceBrackets:

    def is_balanced(self, expression: str) -> bool:
        bracket_map = {')': '(', '}': '{', ']': '['}
        open_brackets = set(bracket_map.values())
        cls_stack = Stack()
        for char in expression:
            if char in open_brackets:
                cls_stack.push(char)
            elif char in bracket_map:
                if cls_stack.is_empty or cls_stack.pop() != bracket_map[char]:
                    return False
        return cls_stack.is_empty



import pandas as pd
from datetime import datetime
from typing import Optional, List, Dict, Any
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response, request
from time import sleep
from zipfile import ZipFile
from stpstone._config.global_slots import YAML_B3_SEARCH_BY_TRADING_SESSION
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.parsers.folders import RemoteFiles
from stpstone.utils.parsers.object import HandlingObjects
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.str import StrHandler


class SearchByTradingB3(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None,
        list_slugs:Optional[List[str]]=None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_B3_SEARCH_BY_TRADING_SESSION,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token,
        self.list_slugs = list_slugs
        self.dt_ref_yymmdd = self.dt_ref.strftime("%y%m%d")
        self.dt_inf_month = DatesBR().dates_inf_sup_month(self.dt_ref)[0]
        self.dt_inf_month_yymmdd = self.dt_inf_month.strftime("%y%m%d")

    def instruments_register_raw(
        self, 
        method_request: str = "GET", 
        key_token: str = "token"
    ) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # api consulta lotes padrão xp
        url_token = "https://arquivos.b3.com.br/api/download/requestname?fileName=InstrumentsConsolidated&date={}".format(
            DatesBR().curr_date)
        url_lotes_padrao_b3 = "https://arquivos.b3.com.br/api/download/?token={}"
        # coletando token para consulta da API da B3
        token_api_b3 = \
            HandlingObjects().literal_eval_data(
                request(method_request, url_token).text)[key_token]
        url_lotes_padrao_b3 = str(url_lotes_padrao_b3).format(token_api_b3)
        # print("URL LOTE PADRAO B3: {}".format(url_lotes_padrao_b3))
        resp_req = request(
            method_request, url_lotes_padrao_b3.format(token_api_b3))
        resp_req.raise_for_status()
        req_resp_text = resp_req.content.decode("iso-8859-1")
        # print("RESPONSE API B3 LOTE PADRAO: {}".format(req_resp_text[:1000]))
        # tratando dados provindos da API
        list_response_api_b3 = req_resp_text.split("\n")
        list_headers = list_response_api_b3[1].split(";")
        # print(f"LIST HEADERS: {list_headers[:10]}")
        list_response_api_b3 = list_response_api_b3[2:]
        # print(f"LIST RESPONSE API B3: {list_response_api_b3[:10]}")
        list_ser = list()
        for row in list_response_api_b3:
            list_row = row.split(";")
            list_ser.append(dict(zip(list_headers, list_row)))
        # criando dataframe à partir do dicionário de interesse
        df_instruments_register_b3_raw = pd.DataFrame(list_ser)
        # changing column types
        list_cols_df = list(df_instruments_register_b3_raw.columns)
        df_instruments_register_b3_raw = df_instruments_register_b3_raw.astype(
            dict(zip(list_cols_df, [str] * len(list_cols_df))))
        # retornando dicionário com dados cadastrais de instrumentos negociados na b3
        return df_instruments_register_b3_raw

    def security_category_name(
        self, 
        dict_export: Dict[Any] = dict(), 
        key_ticker: str = "TckrSymb", 
        col_security_category_name: str = "SctyCtgyNm", 
        col_market_name: str = "MktNm", 
        list_markets_classified: List[str] = list(), 
        bl_return_markets_not_classified: bool = False
    ) -> Dict[str, List[str]]:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # fetch in memory instruments register of assets traded in b3 exchange
        df_instruments_register_b3_raw = self.instruments_register_raw()
        # print("*** REIGSTER B3 RAW ***")
        # print(df_instruments_register_b3_raw)
        # creating dictionary with instruments according to each type of market
        for security_division, col_ in [
            ("securities_by_category_name", col_security_category_name),
                ("securities_by_market_name", col_market_name)]:
            for market, list_source_names in YAML_B3_SEARCH_BY_TRADING_SESSION["up2data_b3"][
                security_division].items():
                #   validating wheter the market is already in the exporting dictionary and extending
                #       or creating a new list
                if market not in dict_export:
                    dict_export[market] = ListHandler().remove_duplicates(
                        df_instruments_register_b3_raw[
                            df_instruments_register_b3_raw[col_].isin(
                                list_source_names)][key_ticker].tolist())
                else:
                    dict_export[market].extend(ListHandler().remove_duplicates(
                        df_instruments_register_b3_raw[
                            df_instruments_register_b3_raw[col_].isin(
                                list_source_names)][key_ticker].tolist()))
                list_markets_classified.extend(list_source_names)
        # defining markets not classified, if is user"s will
        if bl_return_markets_not_classified == True:
            return ListHandler().remove_duplicates(
                df_instruments_register_b3_raw[
                    ~df_instruments_register_b3_raw[col_security_category_name].isin(
                        list_markets_classified)][col_security_category_name].tolist())
        else:
            #   returing dictionary with tickers accordingly to the market type which participates
            return dict_export

    def carga_mtm_b3(self) -> List[str]:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # url de exportação do arquivo de margens teóricas máximas b3
        url_mtm = "https://www.b3.com.br/pesquisapregao/download?filelist=MT{}.zip".format(
            self.dt_ref.strftime("%y%m%d"))
        # carga para a memória margens teóricas máximas b3
        resp_req = request("GET", url_mtm)
        zipfile = RemoteFiles().get_zip_from_web_in_memory(
            resp_req, bl_io_interpreting=False)
        # retorno é um novo zip com um xlsx
        zipfile = ZipFile(zipfile)
        # coletando nome do arquivo txt contido dentro do zip
        list_zip_files = zipfile.namelist()
        file_name = list_zip_files.pop()
        # extraindo arquivo com fatores de risco primitivos b3 - txt
        name_file_extract = zipfile.open(file_name)
        # lendo csv com margens teóricas máximas b3
        list_rows_mtm = [
            str(str_).replace(r'\n', '').replace(
                "b'", '').replace("'", '').split(';')
            for str_ in name_file_extract.readlines() if len(
                str(str_).replace(r'\n', '').replace("b'", '').replace(
                    "'", '').split(';')) > 1
        ]
        list_rows_mtm = list_rows_mtm[1:]
        # retornando lista de linhas do csv corrente
        return list_rows_mtm

    def mtm_compra_venda(
        self, 
        den_desagios: Optional[float] = None, 
        key_stocks: str = "stocks", 
        key_funds:str = "funds", 
        key_etfs: str = "etfs", 
        key_bdrs: str = "bdrs"
    ) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS
        """
        # setting variables
        list_num_desagios_compra = list()
        list_num_desagios_venda = list()
        list_exportacao = list()
        # carga de margens teóricas máximas crua
        list_rows_mtm = self.carga_mtm_b3()
        # carga de ativos bovespa e bmf por tipo
        dict_ativos_bov_bmf_tipo = self.security_category_name()
        # selecionando apenas lista de ativos de interesse - ações, fiis, etfs, e bdrs
        list_ativos_bov_vista = ListHandler().extend_lists(
            dict_ativos_bov_bmf_tipo[key_stocks], dict_ativos_bov_bmf_tipo[key_funds],
            dict_ativos_bov_bmf_tipo[key_etfs], dict_ativos_bov_bmf_tipo[key_bdrs])
        # removendo eventuais duplicatas
        list_ativos_bov_vista = ListHandler().remove_duplicates(list_ativos_bov_vista)
        # alocando margens teóricas máximas em lista de dicionários
        for i, _ in enumerate(list_rows_mtm):
            if i < len(list_rows_mtm) - 3:
                #   setando variáveis de passagem
                dict_exportacao = dict()
                #   margem B3 requerida para passar posicionado no papel
                if list_rows_mtm[i][0] == "INSTRUMENT" \
                    and list_rows_mtm[i][4] != "0" \
                    and (len(list_rows_mtm[i][5]) == 5 or
                         len(list_rows_mtm[i][5]) == 6):
                    den_desagios = float(
                        list_rows_mtm[i][4].replace(",", "."))
                #   numerador de preço com choque para cálculo de deságio por cenário
                #       dias úteis considerados: 1 a 10 dias úteis
                #       garantias exigidas calculadas de compra e venda
                for dias_uteis_choque in range(1, 11):
                    try:
                        if int(list_rows_mtm[i + dias_uteis_choque][1]) == \
                            dias_uteis_choque \
                                and list_rows_mtm[i + dias_uteis_choque][2] != "":
                            list_num_desagios_compra.append(float(
                                list_rows_mtm[i + dias_uteis_choque][2].replace(",", ".")))
                        if int(list_rows_mtm[i + dias_uteis_choque][
                            1]) == dias_uteis_choque and list_rows_mtm[
                                i + dias_uteis_choque][3] != "":
                            list_num_desagios_venda.append(float(
                                list_rows_mtm[i + dias_uteis_choque][3].replace(",", ".")))
                    except:
                        pass
                # executar o loop para listas vazias
                if (list_num_desagios_compra != []) and (list_num_desagios_venda != []):
                    # denominador com preço spot no fechamento para cálculo do deságio
                    #   B3 por cenário
                    if (list_rows_mtm[i][0] == "INSTRUMENT") and (list_rows_mtm[i][4] != "0") \
                            and all([type(x) == float for x in
                                     list_num_desagios_compra]) and (type(
                            den_desagios) == float):
                        # setando dicionário para o ticker de interesse
                        dict_exportacao["TICKER"] = list_rows_mtm[i][5]
                        # definindo preço de fechamento de referência
                        dict_exportacao["PRECO_FECHAMENTO"] = den_desagios
                        # definindo deságios para cenários de choque para o período de
                        #   interesse
                        for dias_uteis_choque in range(1, 11):
                            # deságio de compra/venda para ativo corrente negociado em mercado
                            #   secundário da b3
                            if list_rows_mtm[i][5] in list_ativos_bov_vista:
                                dict_exportacao["MARGEM_B3_COMPRA_D_" + str(
                                    dias_uteis_choque)] = abs(den_desagios * (
                                        1 - list_num_desagios_compra[
                                            dias_uteis_choque - 1] / den_desagios))
                                dict_exportacao["DESAGIO_B3_COMPRA_D_" + str(
                                    dias_uteis_choque)] = min([1, abs(1 - list_num_desagios_compra[
                                        dias_uteis_choque - 1] / den_desagios)])
                                dict_exportacao["MARGEM_B3_VENDA_D_" + str(
                                    dias_uteis_choque)] = \
                                    abs(den_desagios * (
                                        1 - list_num_desagios_venda[
                                            dias_uteis_choque - 1] / den_desagios))
                                dict_exportacao["DESAGIO_B3_VENDA_D_" + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(1 - list_num_desagios_venda[
                                        dias_uteis_choque - 1] / den_desagios)])
                            else:
                                dict_exportacao["MARGEM_B3_COMPRA_D_" + str(
                                    dias_uteis_choque)] = abs(
                                    -list_num_desagios_compra[
                                        dias_uteis_choque - 1])
                                dict_exportacao["DESAGIO_B3_COMPRA_D_" + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(list_num_desagios_compra[
                                        dias_uteis_choque - 1] / den_desagios - 1)])
                                dict_exportacao["MARGEM_B3_VENDA_D_" + str(
                                    dias_uteis_choque)] = \
                                    abs(-list_num_desagios_venda[dias_uteis_choque - 1])
                                dict_exportacao["DESAGIO_B3_VENDA_D_" + str(
                                    dias_uteis_choque)] = \
                                    min([1, abs(list_num_desagios_venda[
                                        dias_uteis_choque - 1] / den_desagios - 1)])
                    # appendando à lista de exportação
                    list_exportacao.append(dict_exportacao)
                    # ressetiando variável de passagem
                    list_num_desagios_compra = list()
                    list_num_desagios_venda = list()
        # appendando lista de exportação a um dataframe
        df_mtm_b3 = pd.DataFrame(list_exportacao)
        # drop not available rows
        df_mtm_b3.dropna(inplace=True)
        # sort tickers in alphabetic order
        df_mtm_b3.sort_values(["TICKER"], ascending=[True], inplace=True)
        # print(f"MTM B3: \n{df_mtm_b3}")
        # retornando deságios b3 de interesse
        return df_mtm_b3

    def collateral_accepted_spot_bov_b3(
        self, 
        dict_repalce_str_erros: Dict[str] = {"ETF": ""}, 
        col_ticker: str = "TICKER", 
        col_isin: str = "ISIN", 
        col_limite_qtds: str = "LIMITE_QUANTIDADES", 
        col_mes_arquivo: str = "MES_ARQUIVO", 
        key_limite_qtds: str = "Limite (quantidade)"
    ) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # setting variables
        list_exportacao = list()
        # importing to memory html, in order to catch the url of the collateral accepted by b3,
        #   regarding spot bov
        resp_req = request("GET", "https://www.b3.com.br/pt_br/produtos-e-servicos/compensacao-e-liquidacao/clearing/administracao-de-riscos/garantias/limites-de-renda-variavel/")
        html_parser_acoes_units_etfs_aceitos_garantia_b3 = HtmlHandler().lxml_parser(resp_req)
        url_acoes_units_etfs_aceitos_garantia_b3 = HtmlHandler().lxml_xpath(
            html_parser_acoes_units_etfs_aceitos_garantia_b3,
            '//*[@id="panel1a"]/ul/li[2]/a/@href')[0]
        url_acoes_units_etfs_aceitos_garantia_b3 = StrHandler().get_string_after_substr(
            url_acoes_units_etfs_aceitos_garantia_b3, "data/files")
        # creating complete url
        url_acoes_units_etfs_aceitos_garantia_b3 = \
            "https://www.b3.com.br/" \
            + "data/files" + \
            url_acoes_units_etfs_aceitos_garantia_b3
        # openning zip in memory
        resp_req = request("GET", url_acoes_units_etfs_aceitos_garantia_b3)
        list_unzipped_files = RemoteFiles().get_zip_from_web_in_memory(resp_req)
        if str(type(list_unzipped_files)) != "<class 'list'>":
            list_unzipped_files = [list_unzipped_files]
        # looping within files in zip
        for file_unz in list_unzipped_files[1:]:
            #   creating columns month-file
            str_mes_arquivo_garantias_b3_corrente = StrHandler().get_between(
                str(file_unz), "_", ".xlsx").replace(" ", "").lower()
            #   looping within sheet names
            for nome_plan_ativa in [
                "Ações, BDRs, ETFs e Units",
                "ADR", 
                "DEB"
            ]:
                #   openning dataframe in memory
                reader = pd.read_excel(
                    file_unz, sheet_name=nome_plan_ativa)
                df_coll_acc_spot_bov = pd.DataFrame(reader)
                #   creating serialized list with spot bov collateral accepted
                for _, row in df_coll_acc_spot_bov.iterrows():
                    list_exportacao.append({
                        col_mes_arquivo: str_mes_arquivo_garantias_b3_corrente.upper(),
                        col_ticker: row["Código"],
                        col_isin: row[col_isin],
                        col_limite_qtds: row[key_limite_qtds]
                    })
        # creating dataframe
        df_coll_acc_spot_bov = pd.DataFrame(list_exportacao)
        # removing duplicates
        df_coll_acc_spot_bov[col_ticker] = [StrHandler().replace_all(
            str(x), dict_repalce_str_erros) for x in df_coll_acc_spot_bov[
                col_ticker].tolist()]
        # returning dataframe
        return df_coll_acc_spot_bov

    def risk_scenarios_curve_types(
        self, 
        list_risk_curve_type_scenarios: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # setting variables
        list_ser = list()
        list_cols_scenarios = ["TIPO", "ID_FPR", "ID_CENARIO", "INT_TIPO_CENARIO",
            "DIAS_HOLDING_PERIOD", "DIAS_CORRIDOS_VERTICE", "DIAS_SAQUE_VERTICE", "VALOR_PHI_1",
            "VALOR_PHI_2"]
        # validando tipo de variáveis de interesse
        if list_risk_curve_type_scenarios != None:
            #   convertendo para lista, caso não seja
            if type(list_risk_curve_type_scenarios) != list:
                list_risk_curve_type_scenarios = [list_risk_curve_type_scenarios]
            #   alterando tipo de instâncias
            list_risk_curve_type_scenarios = [str(x)
                                        for x in list_risk_curve_type_scenarios]
        # definindo url com cenários de risco - tipo de curva
        url_risk_curve_type_scenarios = "https://download.bmfbovespa.com.br/FTP/IPNv2/RISCO/CenariosTipoCurva{}.zip".format(
            self.dt_ref.strftime("%y%m%d"))
        resp_req = request("GET", url_risk_curve_type_scenarios)
        resp_req.raise_for_status()
        # baixando em memória zip com cenário de tipo curva da b3
        list_txts_curve_type_scenarios = RemoteFiles().get_zip_from_web_in_memory(
            resp_req, bl_io_interpreting=False)
        # loopando em torno dos arquivos em txt, abrindo em memória no dataframe pandas
        for extracted_file_curve_type_scenarios in list_txts_curve_type_scenarios:
            #   caso a variável list_risk_curve_type_scenarios seja diferente de none, o usuário tem
            #       interesse que apenas uma parte dos cenários seja considerado, com isso, caso
            #       o cenário corrente não esteja nessa amostra passar para o próximo
            if (list_risk_curve_type_scenarios != None) and (all([StrHandler().find_substr_str(
                    extracted_file_curve_type_scenarios.name, substr) == False for substr in
                    list_risk_curve_type_scenarios])):
                continue
            #   abrindo em memória arquivo corrente - pandas dataframe
            reader = pd.read_csv(extracted_file_curve_type_scenarios, sep=";", skiprows=1, 
                                 header=None, names=list_cols_scenarios, decimal=",")
            df_cenarios_passagem = pd.DataFrame(reader)
            #   preenchendo com valor padrão campos com valor indisponível
            df_cenarios_passagem.fillna("-1", inplace=True)
            #   alterando para dicionário dataframe corrente, visando consolidar em uma lista e
            #       importar para uma dataframe para exportação"
            list_ser.extend(df_cenarios_passagem.to_dict(orient="records"))
        # adicionando lista a um dataframe a ser exportado
        df_risk_curve_type_scenarios = pd.DataFrame(list_ser)
        # alterando tipo de variáveis de colunas de interesse
        df_risk_curve_type_scenarios = df_risk_curve_type_scenarios.astype({
            list_cols_scenarios[0]: int,
            list_cols_scenarios[1]: int,
            list_cols_scenarios[2]: int,
            list_cols_scenarios[3]: int,
            list_cols_scenarios[4]: int,
            list_cols_scenarios[5]: int,
            list_cols_scenarios[6]: int,
            list_cols_scenarios[7]: float,
            list_cols_scenarios[8]: float,
        })
        # exportando dataframe de interesse
        return df_risk_curve_type_scenarios

    def scenarios_risk_spot(
        self, 
        list_scenarios_risk_spot: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # setting variables
        list_ser_exportacao_curvas = list()
        list_cols_scenarios = ["TIPO", "ID_FPR", "ID_CENARIO", "INT_TIPO_CENARIO",
            "DIAS_HOLDING_PERIOD", "DIAS_CORRIDOS_VERTICE", "DIAS_SAQUE_VERTICE", "VALOR_PHI_1",
            "VALOR_PHI_2"]
        dict_variaveis_tipo_cenarios_curvas = {
            1: "ENVELOPE",
            2: "COERENTE",
            3: "ZIG ZAG"
        }
        # validando tipo de variáveis de interesse
        if list_scenarios_risk_spot != None:
            #   convertendo para lista, caso não seja
            if type(list_scenarios_risk_spot) != list:
                list_scenarios_risk_spot = [list_scenarios_risk_spot]
            #   alterando tipo de instâncias
            list_scenarios_risk_spot = [str(x) for x in list_scenarios_risk_spot]
        try:
            # definindo url com cenários de risco - tipo de curva
            url_cenarios_tipo_spot = "https://download.bmfbovespa.com.br/FTP/IPNv2/RISCO/CenariosTipoSpot{}.zip".format(
                self.dt_ref.strftime("%y%m%d"))
            resp_req = request("GET", url_cenarios_tipo_spot)
            resp_req.raise_for_status()
            # baixando em memória zip com cenário de tipo curva da b3
            list_txts_tipos_spot_cenarios = RemoteFiles().get_zip_from_web_in_memory(
                resp_req,
                bl_io_interpreting=False
            )
        except:
            url_cenarios_tipo_spot = "https://download.bmfbovespa.com.br/FTP/IPNv2/RISCO/CenariosTipoSpot{}.zip".format(
                self.dt_ref.strftime("%y%m%d"))
            resp_req = request("GET", url_cenarios_tipo_spot)
            resp_req.raise_for_status()
            list_txts_tipos_spot_cenarios = RemoteFiles().get_zip_from_web_in_memory(
                url_cenarios_tipo_spot, bl_verify=False,
                bl_io_interpreting=False
            )
        # loopando em torno dos arquivos em txt, abrindo em memória no dataframe pandas
        for extracted_file_tipo_curva_cenario in list_txts_tipos_spot_cenarios:
            #   caso a variável list_scenarios_risk_spot seja diferente de none, o usuário tem
            #       interesse que apenas uma parte dos cenários seja considerado, com isso, caso
            #       o cenário corrente não esteja nessa amostra passar para o próximo
            if (list_scenarios_risk_spot != None) and (all([StrHandler().find_substr_str(
                extracted_file_tipo_curva_cenario.name, substr) == False for substr in
                list_scenarios_risk_spot])): continue
            #   abrindo em memória arquivo corrente - pandas dataframe
            reader = pd.read_csv(extracted_file_tipo_curva_cenario, sep=";", skiprows=1, 
                                 header=None, names=list_cols_scenarios, decimal=",")
            df_cenarios_passagem = pd.DataFrame(reader)
            #   preenchendo com valor padrão campos com valor indisponível
            df_cenarios_passagem.fillna("-1", inplace=True)
            #   alterando para dicionário dataframe corrente, visando consolidar em uma lista
            #       e importar para uma dataframe para exportação
            list_passagem = df_cenarios_passagem.to_dict(orient="records")
            list_ser_exportacao_curvas.extend(list_passagem)
        # adicionando lista a um dataframe a ser exportado
        df_cenarios_tipo_spot = pd.DataFrame.from_dict(list_ser_exportacao_curvas)
        # alterando tipo de variáveis de colunas de interesse
        df_cenarios_tipo_spot = df_cenarios_tipo_spot.astype({
            list_cols_scenarios[0]: int,
            list_cols_scenarios[1]: int,
            list_cols_scenarios[2]: int,
            list_cols_scenarios[3]: int,
            list_cols_scenarios[4]: int,
            list_cols_scenarios[5]: float,
            list_cols_scenarios[6]: float,
        })
        # adicionando colunas de interesse - nome tipo do cenário
        df_cenarios_tipo_spot["NOME_TIPO_CENARIO"] = \
            [dict_variaveis_tipo_cenarios_curvas[x] for x in
                df_cenarios_tipo_spot[list_cols_scenarios[3]].tolist()]
        # exportando dataframe de interesse
        return df_cenarios_tipo_spot

    def primitive_risk_factors_merged(
        self, 
        list_risk_scenarios_curve_type: List[str] = \
            ["2962", "2906", "2945", "2924", "2944", "2925", "2946", "2901", "2902", "2934"], 
        list_scenarios_risk_spot: List[str] = \
            ["2962", "2906", "2945", "2924", "2944", "2925", "2946", "2901", "2902", "2934"], 
        int_cenario_sup_avaliacao: int = 529, 
        bl_debug: bool = False
    ) -> pd.DataFrame:
        """
        DOCSTRING: UNINDO FATORES PRIMITIVOS DE RISCO, CURVAS SPOT E POR TIPO EM UM DATAFRAME COM
            OS IDS DE CENÁRIOS DE ITNERESSE
        INPUTS: -
        OUTPUTS: DATAFRAME
        """
        dict_prfs = {
            "ID_FPR": ["2962", "2906", "2945", "2924", "2944", "2925", "2946", "2901", "2902", "2934"],
            "TIPO_FPR": ["Curva", "Spot", "Curva", "Spot", "Curva", "Spot", "Curva", "Spot", "Spot", "Spot"]
        }
        dict_fprs = {
            2902: "ACC_IPCA",
            2901: "ACC_IGPM",
            2946: "C_IGP",
            2944: "CIPCA",
            2945: "CUPOM",
            2906: "DOLAR",
            2924: "IDI",
            2925: "IGP",
            2962: "PRE",
            2934: "PRTIP"
        }
        # criando dataframe de tipos de frps de interesse
        df_tipos_fprs = pd.DataFrame(dict_prfs)
        # convertendo tipo de colunas de interesse
        df_tipos_fprs = df_tipos_fprs.astype({
            list(dict_prfs.keys())[0]: int,
            list(dict_prfs.keys())[1]: str,
        })
        if bl_debug == True:
            print("*** TIPO FRPS ***")
            print(df_tipos_fprs)
        # * arquivos de pregão b3 - cenários tipo curva
        df_risk_scenarios_curve_type = self.risk_scenarios_curve_types(list_risk_scenarios_curve_type)
        # removendo colunas de interesse
        df_risk_scenarios_curve_type.drop(["TIPO", "VALOR_PHI_2"], axis=1, inplace=True)
        # limitando cenários de tipos de curvas dentro do range de interesse
        df_risk_scenarios_curve_type = df_risk_scenarios_curve_type[
            df_risk_scenarios_curve_type["ID_CENARIO"].isin(range(1, 529))].reset_index(drop=True)
        # eliminando cenários prospectivos
        df_risk_scenarios_curve_type = df_risk_scenarios_curve_type[df_risk_scenarios_curve_type["ID_CENARIO"].isin(
            range(1, int_cenario_sup_avaliacao))].reset_index(drop=True)
        if bl_debug == True:
            print("*** TIPO CURVA ***")
            print(df_risk_scenarios_curve_type.info())
            print(df_risk_scenarios_curve_type)
        # * arquivos de pregão b3 - cenários risco tipo spot
        df_cenarios_tipo_spot = self.scenarios_risk_spot(list_scenarios_risk_spot)
        # removendo colunas de interesse
        df_cenarios_tipo_spot.drop([
            "TIPO",
            "VALOR_PHI_2"
        ], axis=1, inplace=True)
        # limitando cenários de curvas spot
        df_cenarios_tipo_spot = df_cenarios_tipo_spot[df_cenarios_tipo_spot["ID_CENARIO"].isin(
            range(1, 529))].reset_index(drop=True)
        # eliminando cenários prospectivos
        df_cenarios_tipo_spot = df_cenarios_tipo_spot[df_cenarios_tipo_spot["ID_CENARIO"].isin(
            range(1, int_cenario_sup_avaliacao))].reset_index(drop=True)
        if bl_debug == True:
            print("*** TIPO SPOT ***")
            print(df_cenarios_tipo_spot.info())
            print(df_cenarios_tipo_spot)
        # * arquivos de pregão b3 - fatores primitivos de risco
        df_fpr_b3 = self.source("primitive_risk_factors", bl_fetch=True)
        # removendo colunas de interesse
        df_fpr_b3.drop(["TIPO"], axis=1, inplace=True)
        # limitando fprs de interesse para fins de cálculo de swaps
        df_fpr_b3 = df_fpr_b3[df_fpr_b3["ID_FPR"].isin(list(dict_fprs.keys()))]
        if bl_debug == True:
            print("*** TIPO FPR ***")
            print(df_fpr_b3.info())
            print(df_fpr_b3)
        # * início tratamento de dados
        # inner-join fprs com dataframe de tipos de frps
        list_colunas_frps = df_fpr_b3.columns
        if bl_debug == True:
            print("*** COLUNAS FRPS ***")
            print(list_colunas_frps)
        list_colunas_frp = ["ID_FPR", "ID_CENARIO", "TIPO_FPR"] + list_colunas_frps.drop([
            "ID_FPR", "NOME_FPR"]).to_list()
        if bl_debug == True:
            print("*** TIPO FPR B3 ***")
            print(df_fpr_b3)
            print("*** TIPOS FPRS ***")
            print(df_tipos_fprs)
        df_fpr_b3 = df_fpr_b3.merge(
            df_tipos_fprs, how="inner", left_on="ID_FPR", right_on="ID_FPR")
        df_fpr_b3 = df_fpr_b3.reindex(columns=list_colunas_frp)
        # concat cenários tipo curva e spot
        df_risk_scenarios_curve_type = pd.concat([df_risk_scenarios_curve_type, df_cenarios_tipo_spot],
                                           ignore_index=True, sort=False)
        if bl_debug == True:
            print("*** CENÁRIOS TIPO CURVA & SPOT ***")
            print(df_risk_scenarios_curve_type)
        # inner-join fprs com cenarios spot e tipo curvas
        df_fpr_b3 = df_fpr_b3.merge(df_risk_scenarios_curve_type, how="inner", left_on="ID_FPR",
                                    right_on="ID_FPR")
        # preenchendo coluna nome fpr
        df_fpr_b3["NOME_FPR"] = [dict_fprs[x] for x in df_fpr_b3["ID_FPR"]]
        # alterando tipo de colunas de interesse
        df_fpr_b3 = df_fpr_b3.astype({
            "DIAS_CORRIDOS_VERTICE": int,
            "DIAS_HOLDING_PERIOD": int,
            "DIAS_SAQUE_VERTICE": int,
        })
        # limitando colunas de interesse
        df_fpr_b3 = df_fpr_b3[[
            "ID_FPR",
            "NOME_FPR",
            "TIPO_FPR",
            "FORMATO_VARIACAO",
            "ID_GRUPO_FPR",
            "ID_CAMARA_INDICADOR",
            "ID_INSTRUMENTO_INDICADOR",
            # "ORIGEM_INSTURMENTO_INDICADOR",
            "BASE",
            "BASE_INTERPOLACAO",
            "CRITERIO_CAPITALIZACAO",
            "ID_CENARIO_y",
            "INT_TIPO_CENARIO",
            "DIAS_CORRIDOS_VERTICE",
            "DIAS_HOLDING_PERIOD",
            "DIAS_SAQUE_VERTICE",
            "VALOR_PHI_1",
        ]]
        # alterando nome de colunas de interesse
        df_fpr_b3.rename(columns={
            "ID_CENARIO_y": "ID_CENARIO"
        }, inplace=True)
        if bl_debug == True:
            print("*** DF FPRS CONSOLIDADO ***")
            print(df_fpr_b3)
        # retornando fpr completo
        return df_fpr_b3

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        source = self.get_query_params(resp_req.url, "source").lower() \
            if self.get_query_params(resp_req.url, "source") is not None else None
        if source == "mtm_b3":
            return self.mtm_compra_venda()
        elif source == "spot_accepted_collateral_b3":
            return self.collateral_accepted_spot_bov_b3()
        elif source == "risk_scenarios_curve_types":
            return self.risk_scenarios_curve_types()
        elif source == "primitive_risk_factors_merged":
            return self.primitive_risk_factors_merged()
        return None



from logging import Logger
import os
import subprocess
from typing import Any, Dict, List, Optional

import pandas as pd
from psycopg import connect, Connection, Cursor
from psycopg.rows import dict_row
from psycopg.sql import Composed, Identifier, SQL

from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.databases.abc import DatabaseConnection
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.json import JsonFiles


class PostgreSQLDB(DatabaseConnection):
    def __init__(
        self,
        dbname: str,
        user: str,
        password: str,
        host: str,
        port: int,
        str_schema: str = "public",
        logger: Optional[Logger] = None,
    ) -> None:
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.str_schema = str_schema
        self.logger = logger
        self.dict_db_config = {
            "dbname": self.dbname,
            "user": self.user,
            "password": self.password,
            "host": self.host,
            "port": self.port,
        }
        self.conn: Connection = connect(**self.dict_db_config, row_factory=dict_row)
        self.cursor: Cursor = self.conn.cursor()
        self.execute(SQL("SET search_path TO {}").format(Identifier(self.str_schema)))

    def execute(self, str_query: str | Composed) -> None:
        self.cursor.execute(str_query)

    def read(
        self,
        str_query: str,
        dict_type_cols: Optional[Dict[str, Any]] = None,
        list_cols_dt: Optional[List[str]] = None,
        str_fmt_dt: Optional[str] = None,
    ) -> pd.DataFrame:
        # retrieving dataframe
        self.cursor.execute(str_query)
        data = self.cursor.fetchall()
        df_ = pd.DataFrame(data)

        # changing data types
        if all([x is not None for x in [dict_type_cols, list_cols_dt, str_fmt_dt]]):
            df_ = df_.astype(dict_type_cols)
            for col_ in list_cols_dt:
                df_[col_] = [
                    DatesBR().str_date_to_datetime(d, str_fmt_dt) for d in df_[col_]
                ]
        return df_

    def insert(
        self,
        json_data: List[Dict[str, Any]],
        str_table_name: str,
        bl_insert_or_ignore: bool = False,
    ) -> None:
        # validate json, in order to have the same keys
        json_data = JsonFiles().normalize_json_keys(json_data)

        if not json_data:
            return

        columns = json_data[0].keys()
        placeholders = SQL(",").join([SQL("%s") for _ in columns])
        table = Identifier(str_table_name)
        cols = SQL(",").join(map(Identifier, columns))

        if bl_insert_or_ignore:
            query = SQL("""
                INSERT INTO {table} ({cols})
                VALUES ({placeholders})
                ON CONFLICT DO NOTHING
            """).format(table=table, cols=cols, placeholders=placeholders)
        else:
            query = SQL("""
                INSERT INTO {table} ({cols})
                VALUES ({placeholders})
            """).format(table=table, cols=cols, placeholders=placeholders)

        try:
            # Convert each record to a tuple in the same order as columns
            records = [tuple(record[col] for col in columns) for record in json_data]

            # Execute with all records at once
            self.cursor.executemany(query, records)
            self.conn.commit()

            if self.logger is not None:
                CreateLog().info(
                    self.logger,
                    f"Successful commit in db {self.dict_db_config['dbname']} "
                    + f"/ table {str_table_name}.",
                )
        except Exception as e:
            self.conn.rollback()
            self.close()
            if self.logger is not None:
                CreateLog().error(
                    self.logger,
                    "ERROR WHILE INSERTING DATA\n"
                    + f"DB_CONFIG: {self.dict_db_config}\n"
                    + f"TABLE_NAME: {str_table_name}\n"
                    + f"JSON_DATA: {json_data}\n"
                    + f"ERROR_MESSAGE: {e}",
                )
            raise Exception(
                "ERROR WHILE INSERTING DATA\n"
                + f"DB_CONFIG: {self.dict_db_config}\n"
                + f"TABLE_NAME: {str_table_name}\n"
                + f"JSON_DATA: {json_data}\n"
                + f"ERROR_MESSAGE: {e}"
            )

    def close(self) -> None:
        self.conn.close()

    def backup(self, str_backup_dir: str, str_bkp_name: str = None) -> str:
        try:
            os.makedirs(str_backup_dir, exist_ok=True)
            backup_file = os.path.join(str_backup_dir, str_bkp_name)
            env = os.environ.copy()
            env["PGPASSWORD"] = self.password

            command = [
                "pg_dump",
                "-h",
                self.host,
                "-p",
                str(self.port),
                "-U",
                self.user,
                "-F",
                "c",
                "-b",
                "-f",
                backup_file,
                self.dbname,
            ]
            subprocess.run(command, check=True, env=env)
            return f"Backup successful! File saved at: {backup_file}"
        except subprocess.CalledProcessError as e:
            return f"Backup failed: {e}"
        except Exception as e:
            return f"An error occurred: {e}"



### DATABASE LOGGING ###

# pypi.org libs
import socket
from datetime import datetime

import pandas as pd

# local libs
from stpstone._config.global_slots import YAML_GEN
from stpstone.utils.cals.handling_dates import DatesBR


class DBLogs:

    def audit_log(
        self, df_: pd.DataFrame, str_url: str, dt_db_ref: datetime, bl_ts_log_str: bool = True
    ) -> pd.DataFrame:
        """
        Adds audit columns to the DataFrame for logging.

        Args:
            - df_ (pandas.DataFrame): DataFrame to update.
            - str_url (str): URL to insert into the DataFrame.
            - dt_last_update (str): Timestamp of the last update.

        Returns:
            - pd.DataFrame: DataFrame with audit log information.
        """
        df_[YAML_GEN["audit_log_cols"]["url"]] = str_url
        df_[YAML_GEN["audit_log_cols"]["ref_date"]] = dt_db_ref
        if bl_ts_log_str == True:
            df_[YAML_GEN["audit_log_cols"]["log_timestamp"]] = DatesBR().utc_log_ts.strftime(
                "%Y-%m-%d %H:%M:%S.%f%z")
        else:
            df_[YAML_GEN["audit_log_cols"]["log_timestamp"]] = DatesBR().utc_log_ts
        return df_

    def insert_user_info(self, df_: pd.DataFrame, user_id: str) -> pd.DataFrame:
        """
        Inserts user information into the dataframe for traceability.

        Args:
            df_ (pandas.DataFrame): DataFrame to update.
            user_id (str): ID of the user responsible for the update.

        Returns:
            pd.DataFrame: Updated DataFrame with user information.
        """
        df_[YAML_GEN["audit_log_cols"]["user"]] = user_id
        return df_

    def insert_host_info(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Inserts the host name (machine) where the data is being processed into the dataframe.

        Args:
            df_ (pandas.DataFrame): DataFrame to update.

        Returns:
            pd.DataFrame: Updated DataFrame with host information.
        """
        df_[YAML_GEN["audit_log_cols"]["host"]] = socket.gethostname()
        return df_

    def insert_error_info(self, df_: pd.DataFrame, error_msg: str) -> pd.DataFrame:
        """
        Inserts error information into the dataframe to track the issues.

        Args:
            df_ (pandas.DataFrame): DataFrame to update.
            error_msg (str): Error message to insert.

        Returns:
            pd.DataFrame: Updated DataFrame with error information.
        """
        df_[YAML_GEN["audit_log_cols"]["error_msg"]] = error_msg
        return df_

    def log_data_insert(
        self, df_: pd.DataFrame, action_type: str, dt_action: str
    ) -> pd.DataFrame:
        """
        Inserts logging information for data insertions into the dataframe.

        Args:
            df_ (pandas.DataFrame): DataFrame to update.
            action_type (str): Type of action (insert, update, delete).
            dt_action (str): Timestamp of the action.

        Returns:
            pd.DataFrame: Updated DataFrame with logging information.
        """
        df_[YAML_GEN["audit_log_cols"]["action_type"]] = action_type
        df_[YAML_GEN["audit_log_cols"]["action_timestamp"]] = dt_action
        return df_



import time
from typing import Dict, Any, Union, List
from urllib3.util import Retry
from requests import Session
from requests.adapters import HTTPAdapter


class ProxyTester:

    def __init__(self, ip: str, port: int) -> None:
        self.ip = ip
        self.port = port
        self.list_status_forcelist: List[int] = [429, 500, 502, 503, 504]

    def _configure_session(self, dict_proxy:Union[Dict[str, str], None]=None,
                            int_retries:int=10, int_backoff_factor:int=1) -> Session:
            retry_strategy = Retry(
                total=int_retries,
                backoff_factor=int_backoff_factor,
                status_forcelist=self.list_status_forcelist
            )
            adapter = HTTPAdapter(max_retries=retry_strategy)
            session = Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
            if dict_proxy is not None:
                session.proxies.update(dict_proxy)
            return session

    @property
    def test_specific_proxy(self, test_url: str = "https://lumtest.com/myip.json") \
        -> Dict[str, Any]:
        """
        Tests a specific proxy IP and port combination

        Args:
            ip: Proxy IP address to test
            port: Proxy port to test
            test_url: URL to use for testing (default checks IP)

        Returns:
            Dictionary with test results containing:
            - success: bool indicating if proxy worked
            - response_time: float with response time in seconds
            - error: str with error message if failed
            - response: dict with response data if successful
        """
        result = {
            'success': False,
            'response_time': None,
            'error': None,
            'response': None
        }
        start_time = time.time()
        try:
            session = self._configure_session(
                dict_proxy={
                    "http": f"http://{self.ip}:{self.port}",
                    "https": f"http://{self.ip}:{self.port}"
                },
                int_retries=0,
                int_backoff_factor=0
            )
            response = session.get(
                test_url,
                headers={
                    "accept": "application/json",
                    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36"
                },
                timeout=(5, 10)
            )
            response.raise_for_status()
            result.update({
                'success': True,
                'response_time': time.time() - start_time,
                'response': response.json()
            })
        except Exception as e:
            result['error'] = str(e)
        return result



from abc import ABC, abstractmethod
from typing import Dict, List, Any, Optional, Protocol, runtime_checkable
import pandas as pd
from logging import Logger
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from stpstone.utils.loggs.create_logs import CreateLog


@runtime_checkable
class DbCursor(Protocol):
    """Protocol defining a generic database cursor interface."""
    def execute(self, query: str, params: Any = None) -> Any: ...
    def fetchone(self) -> Any: ...
    def fetchall(self) -> List[Any]: ...
    def close(self) -> None: ...


@runtime_checkable
class DbConnection(Protocol):
    """Protocol defining a generic database connection interface."""
    def cursor(self) -> DbCursor: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def close(self) -> None: ...


class DatabaseConnection(ABC, metaclass=TypeChecker):
    """
    Abstract base class for database connections that enforces:
    - self.conn (DbConnection) and self.cursor (DbCursor) must be created in __init__
    - Standard database operations interface
    """
    
    # class variables for type hinting (enforces instance attributes)
    conn: DbConnection
    cursor: DbCursor
    
    @abstractmethod
    def __init__(self, dbname: str, user: str, password: str, host: str, port: int,
                 str_schema: str = 'public', logger: Optional[Logger] = None) -> None:
        """
        Initialize database connection. Concrete implementations MUST:
        1. Create self.conn (DbConnection)
        2. Create self.cursor (DbCursor)
        3. Initialize other necessary attributes
        
        Parameters
        ----------
        dbname : str
            Database name
        user : str
            Database user
        password : str
            Database password
        host : str
            Database host
        port : int
            Database port
        str_schema : str, optional
            Schema name, defaults to 'public'
        logger : Optional[Logger], optional
            Logger instance, defaults to None
        """
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.str_schema = str_schema
        self.logger = logger

    @abstractmethod
    def execute(self, str_query: str) -> None:
        """
        Execute a SQL query without returning results.
        
        Parameters
        ----------
        str_query : str
            SQL query to execute
        """
        pass

    @abstractmethod
    def read(self, str_query: str, 
             dict_type_cols: Optional[Dict[str, Any]] = None,
             list_cols_dt: Optional[List[str]] = None, 
             str_fmt_dt: Optional[str] = None) -> pd.DataFrame:
        """
        Execute a query and return results as DataFrame.
        
        Parameters
        ----------
        str_query : str
            SQL query to execute
        dict_type_cols : Optional[Dict[str, Any]], optional
            Column type mapping, defaults to None
        list_cols_dt : Optional[List[str]], optional
            Date columns to parse, defaults to None
        str_fmt_dt : Optional[str], optional
            Date format string, defaults to None
            
        Returns
        -------
        pd.DataFrame
            Query results
        """
        pass

    @abstractmethod
    def insert(self, json_data: List[Dict[str, Any]], 
               str_table_name: str,
               bl_insert_or_ignore: bool = False) -> None:
        """
        Insert data into a table.
        
        Parameters
        ----------
        json_data : List[Dict[str, Any]]]
            Data to insert (list of dicts)
        str_table_name : str
            Target table name
        bl_insert_or_ignore : bool, optional
            If True, ignore duplicates, defaults to False
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the database connection."""
        pass

    @abstractmethod
    def backup(self, str_backup_dir: str, 
               str_bkp_name: Optional[str] = None) -> str:
        """
        Create database backup.
        
        Parameters
        ----------
        str_backup_dir : str
            Backup directory path
        str_bkp_name : Optional[str], optional
            Custom backup filename, defaults to None
            
        Returns
        -------
        str
            Backup status message
        """
        pass

    def __enter__(self):
        """Support context manager protocol."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit handler. Closes the connection and:
        - Returns False/None to let exceptions propagate
        - Logs any errors during closing
        """
        try:
            self.close()
            if exc_type is not None and self.logger is not None:
                CreateLog().error(
                    self.logger,
                    f"Context exited with exception: {exc_type.__name__}: {exc_val}"
                )
        except Exception as e:
            if self.logger is not None:
                CreateLog().error(
                    self.logger,
                    f"Error closing connection: {str(e)}"
                )
            raise
        return None


from heapq import heappush, heappop

class PriorityQueue:

    def __init__(self):
        self._queue = []
        self._index = 0

    def push(self, item, priority):
        heappush(self._queue, (-priority, self._index, item))
        self._index += 1

    def pop(self):
        return heappop(self._queue)[-1]



import backoff
import re
from logging import Logger
from abc import ABC, abstractmethod
from typing import Optional, Dict, List, Any
from requests import request, Response
from requests.exceptions import ReadTimeout, ConnectTimeout, ChunkedEncodingError, ConnectionError
from lxml import html
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.loggs.create_logs import CreateLog


class AnbimaDataUtils:

    def __init__(self, dict_metadata: Dict[str, str], cls_db: Optional[Any] = None,
                 bl_schema: bool = True, str_tbl_name: Optional[str] = None,
                 str_schema_name: Optional[str] = None, bl_insert_or_ignore: Optional[bool] = False,
                 logger: Optional[Logger] = None) -> None:
        self.dict_metadata  = dict_metadata
        self.cls_db = cls_db
        self.bl_schema = bl_schema
        self.str_tbl_name = str_tbl_name
        self.str_schema_name = str_schema_name
        self.bl_insert_or_ignore = bl_insert_or_ignore
        self.logger = logger

    def _log_info(self, message: str):
        if self.logger is not None:
            CreateLog().info(self.logger, message)
        else:
            print(f"INFO: {message}")

    def get_property(self, property_: str, resource: Optional[str] = None) -> str:
        if resource is not None:
            return self.dict_metadata[resource].get(property_, None) \
                if self.dict_metadata[resource].get(property_, None) is not None \
                else self.dict_metadata["credentials"].get(property_, None)
        else:
            return self.dict_metadata["credentials"].get(property_, None)

    def insert_db(self, list_ser) -> None:
        if self.bl_schema == False:
            str_table_name = f"{self.str_schema_name}_{self.str_tbl_name}"
        self.cls_db.insert(
            list_ser,
            str_table_name=str_table_name,
            bl_insert_or_ignore=self.bl_insert_or_ignore,
        )


class AnbimaDataDecrypt(AnbimaDataUtils):

    def __init__(self, dict_metadata: Dict[str, str], cls_db: Optional[Any] = None,
                 bl_schema: bool = True, str_tbl_name: Optional[str] = None,
                 str_schema_name: Optional[str] = None,
                 bl_insert_or_ignore: Optional[bool] = False) -> None:
        self.dict_metadata  = dict_metadata
        self.cls_db = cls_db
        self.bl_schema = bl_schema
        self.str_tbl_name = str_tbl_name
        self.str_schema_name = str_schema_name
        self.bl_insert_or_ignore = bl_insert_or_ignore

    def urls_funds_builder(
        self,
        int_lower_bound: int = 0,
        int_upper_bound: int = 1_000_000,
        int_step: int = 1,
        str_prefix: str = "C",
        int_length: int = 11,
        list_ids_ignore: Optional[List[str]] = None,
        fstr_url: str = "https://data.anbima.com.br/fundos/{}/indicadores"
    ) -> List[Dict[str, str]]:
        list_ser = list()
        for i in range(int_lower_bound, int_upper_bound, int_step):
            id_ = StrHandler().fill_zeros(str_prefix, i, int_length)
            if id_ in list_ids_ignore: continue
            url = fstr_url.format(id_)
            resp_req = request("GET", url, headers=self.get_property("headers"),
                               cookies=self.get_property("cookies"))
            list_ser.append({
                "COD_ANBIMA": id_,
                "URL": url,
                "STATUS_CODE": resp_req.status_code
            })
        if self.cls_db is not None:
            self.insert_db(list_ser)
        return list_ser


class AnbimaDataFetcher(AnbimaDataUtils):

    def __init__(self, resource: str, dict_metadata: Dict[str, Any], list_slugs: List[str],
                 str_bucket_name: str, session: YieldFreeProxy, client_s3: Any,
                 logger: Optional[Logger] = None) -> None:
        self.resource = resource
        self.dict_metadata = dict_metadata
        self.list_slugs = list_slugs
        self.str_bucket_name = str_bucket_name
        self.session = session
        self.client_s3 = client_s3
        self.logger = logger

    @backoff.on_exception(
        backoff.expo,
        (ReadTimeout, ConnectTimeout, ChunkedEncodingError, ConnectionError),
        max_tries=20,
        base=2,
        factor=2,
        max_value=1200
    )
    def _req_wo_session(self, url: str) -> Response:
        return request(
            "GET",
            url,
            verify=self.get_property("verify", self.resource),
            headers=self.get_property("headers", self.resource),
            params=self.get_property("params", self.resource),
            cookies=self.get_property("cookies", self.resource),
        )

    def _get_data(self, slug: str) -> Optional[html.HtmlElement]:
        app_ = StrHandler().fill_placeholders(
            self.dict_metadata[self.resource]["app"], {"slug": slug})
        url = self.dict_metadata["credentials"]["host"] + app_
        if self.session is None:
            resp_req = self._req_wo_session(url)
        else:
            resp_req = self.session.get(
                url,
                verify=self.get_property("verify", self.resource),
                headers=self.get_property("headers", self.resource),
                params=self.get_property("params", self.resource),
                cookies=self.get_property("cookies", self.resource),
            )
        if resp_req.status_code == 200:
            return None
        return HtmlHandler().lxml_parser(resp_req)

    @property
    def filtered_slugs(self) -> List[str]:
        list_slugs_stored = [str(x).replace(".html", "") for x in
                             self.client_s3.list_objects(self.str_bucket_name)]
        return [x for x in self.list_slugs if x not in list_slugs_stored]

    @property
    def store_s3_data(self) -> None:
        for slug in self.filtered_slugs:
            html_content = self._get_data(slug)
            if html_content is not None:
                html_str = html.tostring(html_content, encoding="unicode", pretty_print=True)
                html_bytes = html_str.encode("utf-8")
                blame_s3 = self.client_s3.put_object_from_bytes(
                    bucket_name=self.str_bucket_name,
                    object_name=f"{slug}.html",
                    data=html_bytes,
                    content_type="text/html"
                )
                self._log_info(f"Status Put Object - Bucket {self.str_bucket_name}: {blame_s3}")
            else:
                self._log_info(f"Empty content, process continued without put action - ID: {slug}")


class AnbimaDataTrt(ABC):

    def __init__(self, url: str, html_content: html.HtmlElement, xpath_script: str,
                 dict_re_patters: Dict[str, str]) -> None:
        self.url = url
        self.html_content = html_content
        self.xpath_script = xpath_script
        self.dict_re_patterns = dict_re_patters

    @property
    def get_re_matches(self) -> Dict[str, str]:
        dict_re_matches = dict()
        for el_ in HtmlHandler().lxml_xpath(self.html_content, self.xpath_script):
            for key, re_pattern in self.dict_re_patterns.items():
                if key not in dict_re_matches: dict_re_matches[key] = list()
                regex_match = re.search(re_pattern, el_)
                if regex_match is not None:
                    for i_match in range(0, len(regex_match.groups()) + 1):
                        regex_group = regex_match.group(i_match).replace('\n', ' ')
                        regex_group = re.sub(r'\s+', ' ', regex_group).strip()
                        regex_group = regex_group.replace('\\', '').replace(r'\"', '"')
                        if regex_group not in dict_re_matches[key]: \
                            dict_re_matches[key].append(regex_group)
            dict_re_matches["url"] = self.url
        return dict_re_matches

    @abstractmethod
    def parse_info(self):
        pass



### SENDING EMAIL THROUGH OUTLOOK CLIENT ###

import os
import win32com.client as win32
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.json import JsonFiles


class DealingOutlook:

    def send_email(self, mail_subject, mail_to=None, mail_cc=None, mail_bcc=None, mail_body=None,
                   mail_attachments=None, send_behalf_of=None, auto_send_email=False,
                   bl_frame_sender=False, bl_read_receipt=False, bl_delivery_receipt=False,
                   bl_image_display=False, str_html_signature=None):
        """
        REFERENCES: https://stackoverflow.com/questions/63435690/send-outlook-email-with-attachment-to-list-of-users-in-excel-with-python
        DOCSTRING: SEND EMAIL WITH ATTACHMENTS AND ON BEHALF OF AN EMAIL, INSTEAD OF THE DEFAULT
        INPUTS: TO, CC, BCC (2 OUT OF THREE ARE OPTIONAL), SUBJECT, BODY, ATTACHMENTS, SEND ON
            BEHALF OF AND AUTO SEND (Y/N)
        OUTPUTS: -
        """
        # outlook object through win32com
        outlook = win32.Dispatch('outlook.application')
        # create mail item
        mail = outlook.CreateItem(0)
        # to
        if mail_to:
            mail.To = mail_to
        # cc
        if mail_cc:
            mail.CC = mail_cc
        # blind copy
        if mail_bcc:
            mail.BCC = mail_bcc
        # subject
        mail.Subject = mail_subject
        # creating send on behalf-of
        if send_behalf_of:
            #   if the frame email is false the send on behalf of ought change the sender account,
            #       nonetheless it will use the main account with a frame to send the email
            if bl_frame_sender == False:
                for account in outlook.Session.Accounts:
                    if str(account) == send_behalf_of:
                        send_behalf_of_account = account
                        break
                mail._oleobj_.Invoke(*(64209, 0, 8, 0, send_behalf_of_account))
            mail.SentOnBehalfOfName = send_behalf_of
        # mail body if not none
        if mail_body != None:
            #   defining mail body
            mail.HTMLBody = mail_body
            #   loading html signature and appending to html body
            if str_html_signature != None:
                mail.HTMLBody += str_html_signature
        # mail attachments
        if mail_attachments != None:
            for mail_attachment in mail_attachments:
                if os.path.exists(mail_attachment):
                    mail.Attachments.Add(Source=mail_attachment)
        # request read receipt
        mail.ReadReceiptRequested = bl_read_receipt
        # request delivery receipt
        mail.OriginatorDeliveryReportRequested = bl_delivery_receipt
        # display created email
        if auto_send_email == False:
            mail.Display()
        elif auto_send_email == 'Y':
            if bl_image_display:
                try:
                    mail.Display()
                    mail.Send()
                except:
                    pass
            else:
                try:
                    mail.Send()
                except:
                    pass
        else:
            mail.Close

    def download_attch(self, email_account, outlook_folder,
                       subj_sub_string, attch_save_path, bl_save_file_w_original_name=False,
                       list_fileformat=None, outlook_subfolder=None, move_to_folder=None,
                       save_only_first_event=False, bl_leave_after_first_occurance=False,
                       bl_break_loops=False):
        """
        DOCTRING: DOWNLOAD A FILE FROM AN SPECIFIC EMAIL
        INPUTS: EMAIL ACCOUNT, OUTLOOK FOLDER, PART OF SUBJECT SUB STRING, ATTACHMENT SAVING PATH
            MOVE TO AN ESPECIFIC FOLDER, SAVE ONLY FIRST EVENT (FALSE AS DEFAULT)
        OUTPUTS: JSON WITH SAVING STATUS (OK, NOK)
        """
        # defining variables
        out_app = win32.Dispatch('Outlook.Application')
        out_namespace = out_app.GetNamespace('MAPI')
        if outlook_subfolder:
            out_iter_folder = out_namespace.Folders[email_account].Folders[
                outlook_folder].Folders[outlook_subfolder]
        else:
            out_iter_folder = out_namespace.Folders[email_account].Folders[outlook_folder]
        if move_to_folder != None:
            out_move_to_folder = out_namespace.Folders[email_account].Folders[move_to_folder]
        dict_attch_saving_status = dict()
        # defining attachments generator
        if str(type(attch_save_path)) == "<class 'list'>":
            list_attch_save_path = [attch for attch in attch_save_path]
        elif str(type(attch_save_path)) == "<class 'str'>":
            list_attach_save_path = list()
            list_attach_save_path.append(attch_save_path)
            list_attch_save_path = [attch for attch in list_attach_save_path]
        else:
            raise Exception(
                'Attachments saving paths ought be a string or list of strings, please check '
                + 'wheter variable type is valid.')
        # counting all itemns in the sub-folder
        item_count = out_iter_folder.Items.Count
        if item_count > 0:
            for i in range(item_count):
                # in case user wants to download only the first occurance, leave
                if bl_break_loops == True:
                    break
                # defining com object of the current message
                message = out_iter_folder.Items[i]
                # find desired mail item and downloanding its attchments
                if StrHandler().find_substr_str(message.Subject, subj_sub_string):
                    for attch_save_path in list_attch_save_path:
                        # in case user wants to download only the first occurance, leave
                        if bl_break_loops == True:
                            break
                        for attch in message.Attachments:
                            # if save file with original name is true, and the file format is the
                            #   desired one, save it in the local source
                            if bl_save_file_w_original_name == True and \
                                DirFilesManagement().get_file_format_from_file_name(
                                    attch.FileName) in list_fileformat:
                                attch.SaveAsFile(
                                    attch_save_path + attch.FileName)
                                dict_attch_saving_status[attch_save_path + attch.FileName] = \
                                    DirFilesManagement().object_exists(
                                        attch_save_path + attch.FileName)
                            # elif complete path to save file is provided, procceed with saving to
                            #   to local source
                            elif bl_save_file_w_original_name == False:
                                attch.SaveAsFile(attch_save_path)
                                dict_attch_saving_status[attch_save_path] = \
                                    DirFilesManagement().object_exists(attch_save_path)
                            # check wheter is important to move email to another folder
                            if move_to_folder != None:
                                message.Move(out_move_to_folder)
                            # check whether is important to save just the first attachment or not
                            if save_only_first_event == True and \
                                    bool(dict_attch_saving_status) == True:
                                return JsonFiles().send_json(dict_attch_saving_status)
                        # in case user wants to download only the first occurance, leave
                        if bl_leave_after_first_occurance == True:
                            bl_break_loops = True
        # return infos with respect to downloaded attachments
        return JsonFiles().send_json(dict_attch_saving_status)

    def received_email_subject_w_rule(self, email_account, outlook_folder,
                                      subj_sub_string, outlook_subfolder=None,
                                      output='subject'):
        """
        REFERENCES: https://stackoverflow.com/questions/22813814/clearly-documented-reading-of-emails-functionality-with-python-win32com-outlook,
            https://docs.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.mailitem?redirectedfrom=MSDN&view=outlook-pia#properties_
        DOCSTRING: LAST EMAIL SUBJECT WITH RULE (OR NAME LIKE)
        INPUTS: EMAIL ACCOUNT, OUTLOOK FOLDER, SUBJECT SUBSTRING, OUTLOOK FOLDER (NONE AS DEFAULT)
        OUTPUTS: OBJECT
        """
        # defining variables
        out_app = win32.Dispatch('Outlook.Application')
        out_namespace = out_app.GetNamespace('MAPI')
        if outlook_subfolder:
            out_iter_folder = out_namespace.Folders[email_account].Folders[
                outlook_folder].Folders[outlook_subfolder]
        else:
            out_iter_folder = out_namespace.Folders[email_account].Folders[outlook_folder]
        # counting all itemns in the sub-folder
        item_count = out_iter_folder.Items.Count
        # list emails subjects
        list_emails_subjects = list()
        if item_count > 0:
            for i in range(item_count):
                message = out_iter_folder.Items[i]
                # find desireed mail item and downloanding its attchments
                if StrHandler().find_substr_str(message.Subject, subj_sub_string):
                    if output == 'subject':
                        list_emails_subjects.append(message.Subject)
                    elif output == 'message_raw':
                        list_emails_subjects.append(message)
                    elif output == 'properties':
                        list_emails_subjects.append({
                            'subject': message.Subject,
                            'last_edition': message.LastModificationTime,
                            'creation_time': message.CreationTime,
                        })
                    else:
                        raise NameError(f'Output {output} is invalid')
        return list_emails_subjects

    def get_body_content(self, email_account, outlook_folder,
                         subj_sub_string, outlook_subfolder=None):
        """
        REFERENCES: https://stackoverflow.com/questions/22813814/clearly-documented-reading-of-emails-functionality-with-python-win32com-outlook,
            https://docs.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.mailitem?redirectedfrom=MSDN&view=outlook-pia#properties_
        DOCSTRING: GET BODY CONTENT
        INPUTS: EMAIL ACCOUNT, OUTLOOK FOLDER, SUBJECT SUBSTRING, OUTLOOK FOLDER (NONE AS DEFAULT)
        OUTPUTS: OBJECT
        """
        # defining variables
        out_app = win32.Dispatch('Outlook.Application')
        out_namespace = out_app.GetNamespace('MAPI')
        if outlook_subfolder:
            out_iter_folder = out_namespace.Folders[email_account].Folders[
                outlook_folder].Folders[outlook_subfolder]
        else:
            out_iter_folder = out_namespace.Folders[email_account].Folders[outlook_folder]
        # counting all itemns in the sub-folder
        item_count = out_iter_folder.Items.Count
        # list emails subjects
        list_emails_subjects = list()
        if item_count > 0:
            for i in range(item_count):
                message = out_iter_folder.Items[i]
                # find desireed mail item and downloanding its attchments
                if StrHandler().find_substr_str(message.Subject, subj_sub_string):
                    list_emails_subjects.append({
                        'subject': message.Subject,
                        'last_edition': message.LastModificationTime,
                        'creation_time': message.CreationTime,
                        'body': message.body,
                    })
        return list_emails_subjects

    def reply_email(self, email_account, outlook_folder, subj_sub_string, msg_body, mail_cc=None,
                    mail_bcc=None, auto_send_email=False, bl_image_display=False, outlook_subfolder=None):
        """
        DOCSTRING: SEND AN AUTOMATED REPLY
        INPUTS: EMAIL ACCOUNT, OUTLOOK FOLDER, SUBJECT SUBSTRING, MESSAGE BODY, MAIL CC, MAIL BCC,
            AUTO SEND EMAIL, BL IMAGE DISPLAY AND OUTLOOK SUBFOLDER (NONE AS DEFAULT)
        OUTPUTS: OBJECT
        """
        # defining variables
        out_app = win32.Dispatch('Outlook.Application')
        out_namespace = out_app.GetNamespace('MAPI')
        if outlook_subfolder:
            out_iter_folder = out_namespace.Folders[email_account].Folders[
                outlook_folder].Folders[outlook_subfolder]
        else:
            out_iter_folder = out_namespace.Folders[email_account].Folders[outlook_folder]
        # counting all itemns in the sub-folder
        item_count = out_iter_folder.Items.Count
        # iterating through folder
        if item_count > 0:
            for i in range(item_count):
                #   creating message object
                message = out_iter_folder.Items[i]
                #   find desireed mail item and sending a reply
                if message.Subject == subj_sub_string:
                    #   reply object
                    reply = message.Reply()
                    #   creating new body
                    reply.HTMLBody = msg_body + reply.HTMLBody
                    #   breaking loop
                    break
        # cc
        if mail_cc:
            reply.CC = mail_cc
        # blind copy
        if mail_bcc:
            reply.BCC = mail_bcc
        # display created email
        if auto_send_email == False:
            reply.Display()
        elif auto_send_email == 'Y':
            if bl_image_display:
                try:
                    reply.Display()
                    reply.Send()
                except:
                    pass
            else:
                try:
                    reply.Send()
                except:
                    pass
        else:
            reply.Close



import pandas as pd
from typing import Optional
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_B3_UP2DATA_REGISTRIES
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class BondIssuersWB3(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None
    ) -> None:
        self.token = self.access_token
        super().__init__(
            dict_metadata=YAML_B3_UP2DATA_REGISTRIES,
            session=session,
            dict_headers=None,
            dict_payload=None,
            cls_db=cls_db,
            logger=logger
        )

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None

    # ! TODO: downstream processing to standardize issuer name in both banks_rts_br and b3_bond_issuers_accp_warranty
    # ! TODO: inner-join bond issuers accepted by b3 with banks participating in brazillian rts



from datetime import date, datetime, timedelta, timezone
from datetime import time as datetime_time
import locale
import time
from typing import List, Optional, Tuple, Union
from zoneinfo import ZoneInfo

import businesstimedelta
from dateutil.relativedelta import relativedelta
from more_itertools import unique_everseen
import pandas as pd
from workalendar.core import SAT, SUN

from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from stpstone.utils.cals.br_bzdays import BrazilBankCalendar
from stpstone.utils.parsers.str import StrHandler


class DatesBR(BrazilBankCalendar, metaclass=TypeChecker):
    def build_date(self, year: int, month: int, day: int) -> date:
        return date(year=year, month=month, day=day)

    def build_datetime(
        self, year: int, month: int, day: int, hour: int, minute: int, second: int
    ) -> datetime:
        return datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )

    def date_to_datetime(
        self, date: date, bl_crop_time: bool = True, bl_tinestamp: bool = True
    ) -> datetime:
        datetime_ = datetime.combine(date, datetime.min.time())
        if bl_tinestamp == True:
            datetime_ = datetime_.timestamp_dt()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def to_integer(self, dt_time: datetime) -> int:
        return 10000 * dt_time.year + 100 * dt_time.month + dt_time.day

    def excel_float_to_date(self, float_excel_date: float) -> date:
        return datetime.fromordinal(date(1900, 1, 1).toordinal() + float_excel_date - 2)

    def excel_float_to_datetime(self, float_excel_date):
        return datetime.fromordinal(
            datetime(1900, 1, 1).toordinal() + float_excel_date - 2
        )

    def check_is_date(self, dt_: datetime) -> bool:
        return isinstance(dt_, date)

    def str_date_to_datetime(
        self, date_str: str, format_input: str = "DD/MM/YYYY"
    ) -> datetime:
        """
        String date to datetime
        Args:
            date_str (str): date in string format
            format_input (str): output format - valid formats: 'DD/MM/YYYY', 'YYYY-MM-DD',
                'YYMMDD', 'DDMMYY', 'DDMMYYYY', 'DD/MM/YY'
        Returns:
            datetime
        """
        if format_input == "DD/MM/YYYY":
            return date(int(date_str[-4:]), int(date_str[3:5]), int(date_str[0:2]))
        elif format_input == "DDMMYY":
            return date(
                int("20" + date_str[-2:]), int(date_str[2:4]), int(date_str[0:2])
            )
        elif format_input == "DDMMYYYY":
            return date(int(date_str[-4:]), int(date_str[2:4]), int(date_str[0:2]))
        elif format_input == "YYYYMMDD":
            return date(int(date_str[:4]), int(date_str[4:6]), int(date_str[6:]))
        elif format_input == "YYYY-MM-DD":
            return date(int(date_str[0:4]), int(date_str[5:7]), int(date_str[-2:]))
        elif format_input == "MM-DD-YYYY":
            return date(int(date_str[-4:]), int(date_str[0:2]), int(date_str[3:5]))
        elif format_input == "YYMMDD":
            return date(
                int("20" + date_str[0:2]), int(date_str[2:4]), int(date_str[-2:])
            )
        elif format_input == "DD/MM/YY":
            return date(
                int("20" + date_str[-2:]), int(date_str[3:5]), int(date_str[0:2])
            )
        elif format_input == "DD.MM.YY":
            return date(
                int("20" + date_str[-2:]), int(date_str[3:5]), int(date_str[0:2])
            )
        else:
            raise Exception(f"Not a valid date format {format_input}")

    def list_wds(
        self, dt_start: datetime, dt_end: datetime, int_wd_bef: int
    ) -> List[int]:
        """
        List of working days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            int_wd_bef (int): number of working days before the start date
        Returns:
            List[int]
        """
        date_ref = self.sub_working_days(self.dt_(), int_wd_bef)
        wd_inf = self.get_working_days_delta(dt_start, date_ref)
        wd_sup = self.get_working_days_delta(dt_end, date_ref)
        return list(range(wd_sup, wd_inf + 1))

    def list_wds(self, dt_start: str, dt_end: str) -> Union[List[datetime], List[str]]:
        """
        List of working days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            List[datetime]
        """
        list_wds = list()
        for x in range(int((dt_end - dt_start).days) + 1):
            list_wds.append(
                super().find_following_working_day(day=dt_start + timedelta(days=x))
            )
        return list(unique_everseen(list_wds))

    def list_cds(self, dt_start: datetime, dt_end: datetime) -> List[datetime]:
        """
        List of calendar days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            format_data (str): format date
        Returns:
            List[datetime]
        """
        list_wds = list()
        for x in range(int((dt_end - dt_start).days)):
            list_wds.append(dt_start + timedelta(days=x))
        return list(unique_everseen(list_wds))

    def list_years(self, dt_start: datetime, dt_end: datetime) -> List[int]:
        """
        List of years between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            List[int]
        """
        list_years = list()
        for x in range(int((dt_end - dt_start).days)):
            list_years.append((dt_start + timedelta(days=x)).year)
        return list(unique_everseen(list_years))

    @property
    def curr_date(self) -> date:
        return date.today()

    @property
    def curr_time(self) -> datetime:
        return datetime.now().time()

    def curr_date_time(
        self, bl_timestamp: bool = False, bl_crop_time: bool = False
    ) -> Union[int, datetime]:
        if bl_timestamp == True:
            datetime_ = datetime.now().timestamp_dt()
        else:
            datetime_ = datetime.now()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def testing_dates(self, dt_start: datetime, dt_end: datetime) -> bool:
        """
        Test if dt_end is greater than dt_start
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            Boolean
        """
        if int((dt_end - dt_start).days) >= 0:
            return True
        else:
            return False

    def year_number(self, dt_: Union[date, datetime]) -> int:
        return int(dt_.strftime("%Y"))

    def day_number(self, dt_: Union[date, datetime]) -> int:
        return int(dt_.strftime("%d"))

    def month_name(
        self,
        dt_: Union[date, datetime],
        bl_abbrv: bool = False,
        local_zone: str = "pt-BR",
    ) -> str:
        """
        Name of the month in the local language
        Args:
            dt_ (date): date
            bl_abbrv (bool): abbreviation
            local_zone (str): local zone
        Returns:
            str
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        if bl_abbrv == True:
            return dt_.strftime("%b")
        else:
            return dt_.strftime("%B")

    def dates_inf_sup_month(self, dt_, last_month_year=12) -> Tuple[date, date]:
        year = self.year_number(dt_)
        month = self.month_number(dt_)
        day = 1
        dt_start = self.find_working_day(self.build_date(year, month, day))
        if month < last_month_year:
            dt_end = self.sub_working_days(self.build_date(year, month + 1, day), 1)
        else:
            dt_end = self.sub_working_days(self.build_date(year + 1, 1, day), 1)
        # returning dates
        return dt_start, dt_end

    def month_number(self, dt_: datetime, bl_month_mm: bool = False) -> Union[int, str]:
        if bl_month_mm == False:
            return int(dt_.strftime("%m"))
        else:
            return dt_.strftime("%m")

    def week_name(
        self, dt_: datetime, bl_abbrv: bool = False, local_zone: str = "pt-BR"
    ) -> str:
        """
        Name of the weekday in the local language
        Args:
            dt_ (date): date
            bl_abbrv (bool): abbreviation
            local_zone (str): local zone
        Returns:
            str
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        if bl_abbrv == True:
            return dt_.strftime("%a")
        else:
            return dt_.strftime("%A")

    def week_number(self, dt_: datetime) -> str:
        return dt_.strftime("%w")

    def find_working_day(self, dt_: datetime) -> datetime:
        return self.add_working_days(self.sub_working_days(dt_, 1), 1)

    def nth_weekday_month(
        self,
        dt_start: datetime,
        dt_end: datetime,
        int_weekday: int,
        nth_rpt: int,
        format_output: str = "DD/MM/YYYY",
        int_days_week: int = 7,
    ) -> List[datetime]:
        """
        Get nth weekday of month
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            int_weekday (int): weekday number
            nth_rpt (int): nth repetition
            format_output (str): format output
            int_days_week (int): number of days in a week
        Returns:
            List[datetime]
        """
        list_wds = self.list_wds(dt_start, dt_end, format_output)
        return [
            self.add_working_days(self.sub_working_days(d, 1), 1)
            for d in list_wds
            if (
                self.week_number(d) == int_weekday
                and d.day >= (nth_rpt * int_days_week - int_days_week)
                and d.day <= (nth_rpt * int_days_week)
            )
        ]

    def delta_calendar_days(self, dt_start: datetime, dt_end: datetime) -> int:
        return (dt_end - dt_start).days

    def add_months(self, dt_, int_months) -> datetime:
        return dt_ + relativedelta(months=int_months)

    def add_calendar_days(self, original_date, days_to_add):
        return original_date + timedelta(days=days_to_add)

    def delta_working_hours(
        self,
        timestamp_inf: str,
        timestamp_sup: str,
        int_hour_start_office: int = 8,
        int_hour_sup_office: int = 18,
        int_hour_start_lunch: int = 0,
        int_hour_sup_lunch: int = 0,
        list_wds: List[int] = [0, 1, 2, 3, 4],
    ) -> int:
        """
        Calculate the number of working hours between two timestamps
        Args:
            timestamp_inf (str): start timestamp_dt
            timestamp_sup (str): end timestamp_dt
            int_hour_start_office (int): start hour office
            int_hour_sup_office (int): end hour office
            int_hour_start_lunch (int): start hour lunch
            int_hour_sup_lunch (int): end hour lunch
            list_wds (List[int]): list of working days
        Returns:
            int
        References: https://pypi.org/project/businesstimedelta/
        """
        # timestamp_dt convertation to datetime
        y_inf, mt_inf, d_inf = (
            int(timestamp_inf.split(" ")[0].split("-")[0]),
            int(timestamp_inf.split(" ")[0].split("-")[1]),
            int(timestamp_inf.split(" ")[0].split("-")[2]),
        )
        h_inf, m_inf, s_inf = (
            int(timestamp_inf.split(" ")[1].split(":")[0]),
            int(timestamp_inf.split(" ")[1].split(":")[1]),
            int(timestamp_inf.split(" ")[1].split(":")[2]),
        )
        y_sup, mt_sup, d_sup = (
            int(timestamp_sup.split(" ")[0].split("-")[0]),
            int(timestamp_sup.split(" ")[0].split("-")[1]),
            int(timestamp_sup.split(" ")[0].split("-")[2]),
        )
        h_sup, m_sup, s_sup = (
            int(timestamp_sup.split(" ")[1].split(":")[0]),
            int(timestamp_sup.split(" ")[1].split(":")[1]),
            int(timestamp_sup.split(" ")[1].split(":")[2]),
        )
        timestamp_inf = datetime(y_inf, mt_inf, d_inf, h_inf, m_inf, s_inf)
        timestamp_sup = datetime(y_sup, mt_sup, d_sup, h_sup, m_sup, s_sup)
        # dict of holidays
        dict_holidays_raw = dict()
        for y in range(timestamp_inf.year, timestamp_sup.year + 1):
            dict_holidays_raw[y] = self.holidays(y)
        dict_holidays_trt = dict()
        for k, v in dict_holidays_raw.items():
            for t in v:
                dict_holidays_trt[t[0]] = t[1]
        # office hours for working days
        workday = businesstimedelta.WorkDayRule(
            start_time=time(int_hour_start_office),
            end_time=time(int_hour_sup_office),
            list_wds=list_wds,
        )
        lunchbreak = businesstimedelta.LunchTimeRule(
            start_time=time(int_hour_start_lunch),
            end_time=time(int_hour_sup_lunch),
            list_wds=list_wds,
        )
        holidays = businesstimedelta.HolidayRule(dict_holidays_trt)
        businesshrs = businesstimedelta.Rules([workday, lunchbreak, holidays])
        # output
        return businesshrs.difference(timestamp_inf, timestamp_sup).timedelta

    def last_wd_years(self, list_years: List[int]) -> List[datetime]:
        """
        Last days of years
        Args:
            list_years (List[int]): list of years
        Returns:
            List[datetime]
        """
        return [self.sub_working_days(datetime(y + 1, 1, 1), 1) for y in list_years]

    def add_holidays_not_considered_anbima(
        self,
        dt_start: datetime,
        dt_end: datetime,
        list_last_week_year_day,
        local_zone="pt-BR",
        list_holidays_not_considered: List[str] = ["25/01"],
        list_dates_not_considered: List[str] = ["05/03/2025", "18/02/2026"],
        list_non_bzdays_week: List[str] = ["sábado", "domingo"],
    ):
        """
        Add holidays not considered by ANBIMA
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            list_last_week_year_day (List[datetime]): list of last days of years
            local_zone (str): locale zone
            list_holidays_not_considered (List[str]): list of holidays not considered
            list_dates_not_considered (List[str]): list of dates not considered
            list_non_bzdays_week (List[str]): list of non business days of the week
        Returns:
            int
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        return len(
            [
                d
                for d in self.list_calendar_days(dt_start, dt_end)
                if (
                    d.strftime("%d/%m") in list_holidays_not_considered
                    and not self.week_name(d) in list_non_bzdays_week
                    or d in list_last_week_year_day
                    or d.strftime("%d/%m/%Y") in list_dates_not_considered
                )
            ]
        )

    def unix_timestamp_to_datetime(
        self, unix_timestamp: Union[float, int], str_tz: str = "UTC"
    ) -> datetime:
        tz_obj = ZoneInfo("UTC") if str_tz == "UTC" else ZoneInfo(str_tz)
        return datetime.fromtimestamp(unix_timestamp, tz=tz_obj)

    def unix_timestamp_to_date(
        self, unix_timestamp: Union[float, int], str_tz: str = "UTC"
    ) -> datetime:
        tz_obj = ZoneInfo("UTC") if str_tz == "UTC" else ZoneInfo(str_tz)
        return datetime.fromtimestamp(unix_timestamp, tz=tz_obj).date()

    def iso_to_unix_timestamp(self, iso_timestamp: str) -> int:
        dt_ = datetime.fromisoformat(iso_timestamp)
        dt_utc = dt_.astimezone(timezone.utc)
        return dt_utc.timestamp()

    def datetime_to_unix_timestamp(
        self, dt_: Union[date, datetime, datetime_time]
    ) -> int:
        """
        Convert a datetime/date/time object to a Unix timestamp (seconds since epoch).

        Args:
            dt_ (Union[date, datetime, time]): Date, datetime or time object to convert

        Returns:
            int: Unix timestamp (seconds since 1970-01-01 00:00:00 UTC)
        """
        # If input is time, combine with today's date
        if isinstance(dt_, datetime_time):
            dt_ = datetime.combine(date.today(), dt_)
        # If input is date, convert to datetime at midnight
        elif isinstance(dt_, date) and not isinstance(dt_, datetime):
            dt_ = datetime.combine(dt_, datetime_time.min)

        # If datetime is timezone-naive, assume local timezone
        if dt_.tzinfo is None:
            dt_ = dt_.astimezone()  # Convert to local timezone

        # Convert to UTC and return timestamp
        return int(dt_.astimezone(timezone.utc).timestamp())

    def timestamp_to_date(
        self,
        timestamp: Union[str, int],
        substring_datetime: Optional[str] = "T",
        format_output: str = "YYYY-MM-DD",
    ) -> datetime:
        if substring_datetime == None:
            return datetime.fromtimestamp(int(timestamp) / 1000, tz=timezone.utc)
        return self.str_date_to_datetime(
            StrHandler().get_string_until_substr(str(timestamp), substring_datetime),
            format_output,
        )

    def timestamp_to_datetime(
        self, timestamp_dt: datetime, bl_return_from_utc: bool = False
    ) -> Union[datetime, str]:
        if bl_return_from_utc == True:
            return pd.to_datetime(timestamp_dt, unit="s", utc=True).tz_convert(
                ZoneInfo("America/Sao_Paulo")
            )
        else:
            return pd.to_datetime(timestamp_dt, unit="s", utc=True).strftime("%Y%m%d")

    @property
    def current_timestamp_string(self, format_output: str = "%Y%m%d_%H%M%S") -> str:
        return self.curr_date_time().strftime(format_output)

    @property
    def utc_log_ts(self) -> datetime:
        return datetime.now(timezone.utc)

    def utc_from_dt(self, dt_):
        dt_ = datetime.combine(dt_, datetime.min.time())
        return dt_.replace(tzinfo=ZoneInfo("UTC"))

    def month_year_string(
        self,
        dt_: str,
        format_input: str = "%b/%Y",
        format_output: str = "%Y-%m",
        bl_dtbr: bool = True,
    ):
        if bl_dtbr == True:
            month_mapping = {
                "JAN": "01",
                "FEB": "02",
                "MAR": "03",
                "APR": "04",
                "MAY": "05",
                "JUN": "06",
                "JUL": "07",
                "AUG": "08",
                "SEP": "09",
                "OCT": "10",
                "NOV": "11",
                "DEC": "12",
            }
            month_abbr, year = dt_.split("/")
            month = month_mapping[month_abbr.upper()]
            return f"{year}-{month}"
        else:
            return datetime.strptime(dt_, format_input).strftime(format_output)
